/******************************************************************************
Central transition function  for vapoursynth by V.C.Mohan
This ClipB  emerging from center gradually Masks  a stationary ClipA. While emerging
options are for resizing to fit, rotating n turns.
Alternately the ClipA disappears in to the center of  ClipB. While disappearing 
options are for resizing to fit, and rotate n times .
 The clips should have identical lengths, height, width and same pixel type.
 Only RGB, YUV and Grey formats are supported. 
 3 Aug 2015 16 July 2021
 Author V.C. Mohan

  (C) < 2015-2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 

*************************************************************************************************/  
//
//-------------------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap;
	int vtransStart;
	int vtransEnd;
	int numFrames;
	bool emerge;;
	bool resize;
	int nTurns;		
} CentralData;	


// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC centralInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    CentralData *d = (CentralData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
}
//______________________________________________________________________________________________
	
/***************************************************************/	
static const VSFrameRef *VS_CC centralGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
    CentralData *d = (CentralData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		
		int deltay =((bn+1) * height)/(d->overlap+1);
		int deltax = ((bn+1) * width)/(d->overlap+1);
		
	//	deltax= deltax & 0xFFFFFFFC;
	//	deltay= deltay & 0xFFFFFFFC;

		if (deltax==0 || deltay==0)
		{
			vsapi->freeFrame(srcb);

			return srca; // no work to be done
		}		

		int degree= (((bn+1) * d->nTurns * 360) / (d->overlap + 1) ) % 360;

		//if(fi->colorFamily == cmRGB) // only for compat

		//	degree = - degree;

		double reduction = (d->overlap + 1.0) / ( bn + 1.0);

		if( ! d->emerge)

			reduction =  (d->overlap + 1.0)/ (d->overlap - bn );
		
		reduction = d->resize? reduction: 1.0;

		int dstx,dsty, wht,wwd,srcx,srcy ;
		const VSFrameRef *base;	// stationary frame
		const VSFrameRef *src; // changing frame
		if(d->emerge)
		{
					// B emerges from center of A Frame expanding in all directions
			
			dstx =(width - deltax)/2;
			dsty =(height - deltay)/2;
			wht = deltay;
			wwd = deltax;
			srcx=(width - deltax)/2;
			srcy=(height - deltay)/2;
			src = srcb;
			base = srca;			
				
		}

		else	// if(!emerge)
			
		{	//LeftFrame disappears into center of RightFrame
			dstx =(deltax)/2;
			dsty =(deltay)/2;
			wht = height - deltay;
			wwd = width - deltax;
			srcx = (deltax)/2;
			srcy = (deltay)/2;
			src = srca;
			base = srcb;
		}

		if (wwd==0 || wht==0)
		{
			vsapi->freeFrame(src);
			return base; // no work to be done
		}

		int subh[] = {0, 0, 0, 0};
		int subw[] = {0, 0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subw[1] = fi->subSamplingW;
			subh[1] = fi->subSamplingH;
			subw[2] = fi->subSamplingW;
			subh[2] = fi->subSamplingH;
		}
		
		VSFrameRef * dst = vsapi->copyFrame(base, core);

		const uint8_t * srcp[] = {NULL, NULL, NULL, NULL};
		uint8_t * dstp[] = {NULL, NULL, NULL, NULL};
		int stride[] = {0, 0, 0, 0};

		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;

		for(int plane = 0; plane < fi->numPlanes; plane ++)
		{
			srcp[plane] = vsapi->getReadPtr(src, plane);
            stride[plane] = vsapi->getStride(src, plane);
            dstp[plane] = vsapi->getWritePtr(dst, plane);
		
			if(! d->resize && d->nTurns == 0)
			{
				int offset = (dsty >> subh[plane]) * stride[plane] + (dstx >> subw[plane]) * nbytes;
					// copy part of frame as no rotation reqd
				vs_bitblt(dstp[plane] + offset, stride[plane],
							srcp[plane] + offset, stride[plane], 
							nbytes * (wwd >> subw[plane]), wht >> subh[plane]);
			
			}
		}

		if (d->resize || d->nTurns != 0)
		{

			if (fi->sampleType == stInteger && fi->bitsPerSample == 8)
			{			
				const uint8_t *sp = srcp[0];
				int spitch = stride[0] / nbytes;
				uint8_t *dp = dstp[0];

				const uint8_t *s1p = srcp[1];
				int s1pitch = stride[1] / nbytes;
				uint8_t *d1p = dstp[1];
				const uint8_t *s2p = srcp[2];
				uint8_t *d2p = dstp[2];

				if ( d->resize )
				{			
					wht = height;
					wwd = width;
				}
				
				if ( fi->numPlanes == 1)
				{
					// grey color formats
					ResizeRotatePlane(dp, height, width, spitch,
										sp, wht,wwd, spitch,
										degree, reduction);	
				}
				else 
				{
					ResizeRotatePYUV(dp, d1p, d2p,
								height, width, spitch, s1pitch,
								sp, s1p, s2p,
								wht, wwd, spitch, s1pitch,
								subh[1],  subw[1],
								 degree, reduction);
				}

			
			}

			else if (fi->sampleType == stInteger && nbits > 8)
			{
				const uint16_t *sp = (const uint16_t *)srcp[0];
				int spitch = stride[0] / nbytes;
				uint16_t *dp = (uint16_t *)dstp[0];

				const uint16_t *s1p = (const uint16_t *)srcp[1];
				int s1pitch = stride[1] / nbytes;
				uint16_t *d1p = (uint16_t *)dstp[1];
				const uint16_t *s2p = (const uint16_t *)srcp[2];
				uint16_t *d2p = (uint16_t *)dstp[2];

				if ( d->resize )
				{			
					wht = height;
					wwd = width;
				}
				
				if ( fi->numPlanes == 1)
				{
					// grey color formats
					ResizeRotatePlane(dp, height, width, spitch, 
										sp, wht,wwd, spitch,
										degree, reduction);	
				}
				else 
				{
					ResizeRotatePYUV(dp, d1p, d2p,
								height, width, spitch, s1pitch,
								sp, s1p, s2p,
								wht, wwd, spitch, s1pitch,
								subh[1],  subw[1],
								 degree, reduction);
				}			
			}
				
		
			else if (fi->sampleType == stFloat)
			{
				const float *sp = (const float *)srcp[0];
				int spitch = stride[0] / nbytes;
				float *dp = (float *)dstp[0];

				const float *s1p = (const float *)srcp[1];
				int s1pitch = stride[1] / nbytes;
				float *d1p = (float *)dstp[1];
				const float *s2p = (const float *)srcp[2];
				float *d2p = (float *)dstp[2];

				if ( d->resize )
				{			
					wht = height;
					wwd = width;
				}
				
				if ( fi->numPlanes == 1)
				{
					// grey color formats
					ResizeRotatePlane(dp, height, width, spitch,
										sp, wht,wwd, spitch,
										degree, reduction);	
				}
				else 
				{
					ResizeRotatePYUV(dp, d1p, d2p,
								height, width, spitch, s1pitch,
								sp, s1p, s2p,
								wht, wwd, spitch, s1pitch,
								subh[1],  subw[1],
								 degree, reduction);
				}

			
			}
				
		}
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}

	return 0;
}

//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC centralFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    CentralData *d = (CentralData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC centralCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    CentralData d;
    CentralData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Central: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Central: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Central: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.nTurns = int64ToIntS(vsapi->propGetInt(in, "nturns", 0, &err));
	if(err)
		d.nTurns = 0;
	else if(d.nTurns < -d.overlap / 8 || d.nTurns > d.overlap / 8)
	{
		vsapi->setError(out, "Central: nturns must result in atleast 8 turns per turn ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}

	temp = !! vsapi->propGetInt(in, "emerge", 0, &err);
	if(err)
		d.emerge = true;
	else if(temp == 0)
		d.emerge = false;
	else 
		d.emerge = true;

	temp = !! vsapi->propGetInt(in, "resize", 0, &err);
	if(err)
		d.resize = true;
	else if(temp == 0)
		d.resize = false;
	else 
		d.resize = true;
			 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (CentralData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Central", centralInit, centralGetFrame, centralFree, fmParallel, 0, data, core);
}

// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "central", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Central", "clip:clip;clipb:clip;overlap:float:opt;nturns:int:opt;emerge:int:opt;resize:int:opt;", centralCreate, 0, plugin);
}
*/