/******************************************************************************
TransCrumple filter plugin for vapoursynth by V.C.Mohan
The A Clip  unmasks a stationary B Clip by folding or crumpling.or
The B Clip masks  the A clip by unfolding or uncrumpling .
 The clips should have identical lengths, height, width and same pixel type 
 5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  

//----------------------------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap;
	int vtransStart;
	int vtransEnd;
	int numFrames;
	bool emerge;	// fold or unfold
	bool crumple; //  is crumple or fanfold

}CrumpleData;	
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC crumpleInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    CrumpleData *d = (CrumpleData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
}
//______________________________________________________________________________________________

static const VSFrameRef *VS_CC crumpleGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
    CrumpleData *d = (CrumpleData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		
		int dy =((bn+1) * height)/( 4 * (d->overlap+1) );
		int dx = ((bn+1) * width)/( 4 * (d->overlap+1) );

			
		if(dx == 0 || dy == 0)
		{
			vsapi->freeFrame(srcb);
			return srca;
		}
		const VSFrameRef *base;	// stationary frame
		const VSFrameRef *src; // changing frame

		if(d->emerge)
		{
			src = srca;
			base = srcb;
			
		}
		else
		{
			base = srca;
			src = srcb;

			dx = width / 4 - dx;
			dy = height / 4 - dy;
		}

		int vwin = height - 2 * dy;
		int wwin = width - 2 * dx;	
			
		int subH[] = {0, 0, 0};
		int subW[] = {0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subW[1] = fi->subSamplingW;
			subH[1] = fi->subSamplingH;
			subW[2] = fi->subSamplingW;
			subH[2] = fi->subSamplingH;
		}
/*		const uint8_t * srcp[] = {NULL, NULL, NULL};
		uint8_t * dstp[] = {NULL, NULL, NULL};
		int stride[] = {0, 0, 0};
*/
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;				
		VSFrameRef * dst = vsapi->copyFrame(base, core);

					
		if(d->crumple)
		{
			VSFrameRef * work = vsapi->newVideoFrame(fi, width, height, src, core);
			VSFrameRef * work2 = vsapi->newVideoFrame(fi, width, height, src, core);


			for(int plane = 0; plane < fi->numPlanes; plane ++)
			{
				const uint8_t * srcp = vsapi->getReadPtr(src, plane);
				const int stride = vsapi->getStride(src, plane);
				uint8_t *dstp = vsapi->getWritePtr(dst, plane);
				int pitch = stride / kb;
				int ht = vsapi->getFrameHeight(srca, plane);
				int wd = vsapi->getFrameWidth(srca, plane);
				uint8_t *wp = vsapi->getWritePtr(work, plane);
				uint8_t * w2p = vsapi->getWritePtr(work2, plane);

				int v1win = vwin >> subH[plane];
				int w1win = wwin >> subW[plane];
				int d1x = dx >> subW[plane];
				int d1y = dy >> subH[plane];
				
				if( fi->sampleType == stInteger && nb == 8)
				{
			
					VFastResizerPlanar(srcp, wp, ht, wd, v1win , wd, pitch, pitch, false);

					HFastResizerPlanar(wp, w2p + pitch * d1y  + d1x , v1win , wd,
									v1win , w1win , pitch, pitch,false);
					

					for(int h = d1y ; h < ht / 2; h ++)
					{
						float iwin = ((float)((h - d1y) * (wd / 2 - d1x) ) ) / (ht / 2 - d1y);
						float owin = ((float)(h - d1y) * (wd / 2 - 2 * d1x) )/(ht / 2 - d1y);
						int wcenter = (int)(d1x + iwin / 2) & 0xfffffffe;
						int owcenter=(int)(d1x + ((float)(h - d1y) * d1x) / (ht / 2 - d1y) + owin / 2) & 0xfffffffe;

						if(owin>=2)
						{
							ResizeHLinePlanar(w2p + pitch * (h ) + wcenter,
								dstp + pitch * (h) + owcenter,
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2p + pitch * (h) + (wd - wcenter),
								dstp + pitch * (h) + (wd - owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2p + pitch * (ht - 1 - h) + (wcenter),
								dstp + pitch * (ht - 1 - h) + (owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2p + pitch * (ht - 1 - h) + (wd - wcenter),
								dstp + pitch * (ht - 1 - h) + (wd - owcenter),
								(int)owin, iwin / owin);
						}
					}				

					for(int w = d1x; w < wd / 2; w ++)
					{
						float iwin = ((float)( w - d1x) * (ht / 2 - d1y)) / (wd / 2 - d1x);

						float owin=( (float)( w - d1x) * (ht / 2 - 2 * d1y) ) / (wd / 2 - d1x);

						int wcenter = (int)(d1y + iwin / 2);

						int owcenter = (int)(d1y + ((float)( w - d1x) * d1y) / (wd / 2 - d1x) + owin / 2);

						if(owin >= 2)
						{
							ResizeVLinePlanar(w2p + pitch * (wcenter) + w,
								dstp + pitch * (owcenter) + w,
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2p + pitch * (wcenter) + (wd - w),
								dstp + pitch * (owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2p + pitch * (ht - 1 - wcenter) + (w),
								dstp + pitch * (ht - 1 - owcenter) + (w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2p + pitch * (ht - 1 - wcenter) + (wd - w),
								dstp + pitch * (ht - 1 - owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);
						}
					}
				}//  stInteger && nb == 8

				else if( fi->sampleType == stInteger && nb > 8)
				{

					const uint16_t * srcpp = (const uint16_t *)srcp;
					uint16_t *dstpp = (uint16_t *)dstp;
					uint16_t * wpp = (uint16_t *)wp;
					uint16_t * w2pp = (uint16_t *)w2p;					
			
					VFastResizerPlanar(srcpp, wpp, ht, wd, v1win , wd, pitch, pitch, false);

					HFastResizerPlanar(wpp, w2pp + pitch * d1y  + d1x , v1win , wd,
									v1win , w1win , pitch, pitch,false);
					

					for(int h = d1y ; h < ht / 2; h ++)
					{
						float iwin= ((float)((h-d1y)*(wd/2-d1x)) )/(ht/2-d1y);
						float owin= ((float)(h-d1y)*(wd/2-2*d1x) )/(ht/2-d1y);
						int wcenter=(int)(d1x+iwin/2) & 0xfffffffe;
						int owcenter=(int)(d1x+ ((float)(h-d1y)*d1x)/(ht/2-d1y)+owin/2) & 0xfffffffe;

						if(owin>=2)
						{
							ResizeHLinePlanar(w2pp + pitch * (h) + wcenter,
								dstpp + pitch * (h) + owcenter,
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (h) + (wd - wcenter),
								dstpp + pitch * (h) + (wd - owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (ht - 1 - h) + (wcenter),
								dstpp + pitch * (ht - 1 - h) + (owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (ht - 1 - h) + (wd - wcenter),
								dstpp+pitch * (ht - 1 - h) + (wd - owcenter),
								(int)owin, iwin / owin);
						}
					}				

					for(int w = d1x; w < wd / 2; w ++)
					{
						float iwin = ((float)( w - d1x) * (ht / 2 - d1y)) / (wd / 2 - d1x);

						float owin=( (float)( w - d1x) * (ht / 2 - 2 * d1y) ) / (wd / 2 - d1x);

						int wcenter = (int)(d1y + iwin / 2);

						int owcenter = (int)(d1y + ((float)( w - d1x) * d1y) / (wd / 2 - d1x) + owin / 2);

						if(owin >= 2)
						{
							ResizeVLinePlanar(w2pp + pitch * (wcenter) + w,
								dstpp + pitch * (owcenter) + w,
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (wcenter) + (wd - w),
								dstpp + pitch * (owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (ht - 1 - wcenter) + (w),
								dstpp + pitch * (ht - 1 - owcenter) + (w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (ht - 1 - wcenter) + (wd - w),
								dstpp + pitch * (ht - 1 - owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);
						}
					}
				}	// nb > 8

				else if( fi->sampleType == stFloat)
				{
					const float* srcpp = (const float*)srcp;
					float*dstpp = (float*)dstp;
					float* wpp = (float*)wp;
					float* w2pp = (float*)w2p; 
			
					VFastResizerPlanar(srcpp, wpp, ht, wd, v1win , wd, pitch, pitch, false);

					HFastResizerPlanar(wpp, w2pp + pitch * d1y  + d1x , v1win , wd,
									v1win , w1win , pitch, pitch,false);
					

					for(int h = d1y ; h < ht / 2; h ++)
					{
						float iwin= ((float)((h-d1y)*(wd/2-d1x)) )/(ht/2-d1y);
						float owin= ((float)(h-d1y)*(wd/2-2*d1x) )/(ht/2-d1y);
						int wcenter=(int)(d1x+iwin/2) & 0xfffffffe;
						int owcenter=(int)(d1x+ ((float)(h-d1y)*d1x)/(ht/2-d1y)+owin/2) & 0xfffffffe;

						if(owin>=2)
						{
							ResizeHLinePlanar(w2pp + pitch * (h) + wcenter,
								dstpp + pitch * (h) + owcenter,
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (h) + (wd - wcenter),
								dstpp + pitch * (h) + (wd - owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (ht - 1 - h) + (wcenter),
								dstpp + pitch * (ht - 1 - h) + (owcenter),
								(int)owin, iwin / owin);
							ResizeHLinePlanar(w2pp + pitch * (ht - 1 - h) + (wd - wcenter),
								dstpp + pitch * (ht - 1 - h) + (wd - owcenter),
								(int)owin, iwin / owin);
						}
					}				

					for(int w = d1x; w < wd / 2; w ++)
					{
						float iwin = ((float)( w - d1x) * (ht / 2 - d1y)) / (wd / 2 - d1x);

						float owin=( (float)( w - d1x) * (ht / 2 - 2 * d1y) ) / (wd / 2 - d1x);

						int wcenter = (int)(d1y + iwin / 2);

						int owcenter = (int)(d1y + ((float)( w - d1x) * d1y) / (wd / 2 - d1x) + owin / 2);

						if(owin >= 2)
						{
							ResizeVLinePlanar(w2pp + pitch * (wcenter) + w,
								dstpp + pitch * (owcenter) + w,
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (wcenter) + (wd - w),
								dstpp + pitch * (owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (ht - 1 - wcenter) + (w),
								dstpp + pitch * (ht - 1 - owcenter) + (w),
								(int)owin, pitch, pitch, iwin / owin);

							ResizeVLinePlanar(w2pp + pitch * (ht - 1 - wcenter) + (wd - w),
								dstpp + pitch * (ht - 1 - owcenter) + (wd - w),
								(int)owin, pitch, pitch, iwin / owin);
						}
					}
				}	// float
				
				
			}	// plane

		}	// crumple

		else  // if(! d->crumple)	// fanfold
		{
			for(int plane = 0; plane < fi->numPlanes; plane ++)
			{
				const uint8_t * srcp = vsapi->getReadPtr(src, plane);
				const int stride = vsapi->getStride(src, plane);
				uint8_t *dstp = vsapi->getWritePtr(dst, plane);
				int pitch = stride / kb;
				int ht = vsapi->getFrameHeight(srca, plane);
				int wd = vsapi->getFrameWidth(srca, plane);

				int d1x = dx >> subW[plane];
				int d1y = dy >> subH[plane];

				if( fi->sampleType == stInteger && nb == 8)
				{

					for(int w=0; w< wd/2;w++)	
					{
	
									// lower quarter image
						ResizeVLinePlanar(srcp+pitch*((w*ht/wd)/2)+w,
										dstp+pitch*(w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcp+pitch*((w*ht/wd)/2)+(wd-w),
										dstp+pitch*(w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
									// top quarter image
						ResizeVLinePlanar(srcp+pitch*(ht-(w*ht/wd)/2)+w,
										dstp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcp+pitch*(ht-(w*ht/wd)/2)+(wd-w),
										dstp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
					}				

					for(int h=0; h < ht/2; h++)
					{
										// Left quadrant image
						ResizeHLinePlanar(srcp+pitch*h+((h*wd/ht)/2),
											dstp+pitch*h+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcp+pitch*(ht-h)+((h*wd/ht)/2),
											dstp+pitch*(ht-h)+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
										// Right quadrant image
						ResizeHLinePlanar(srcp+pitch*h+(wd-(h*wd/ht)/2),
											dstp+pitch*h+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcp+pitch*(ht-h)+(wd-(h*wd/ht)/2),
											dstp+pitch*(ht-h)+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
					}
				}// int and nb = 8

				else if( fi->sampleType == stInteger && nb > 8)
				{

					const uint16_t * srcpp = (const uint16_t *)srcp;
					uint16_t *dstpp = (uint16_t *)dstp;

					for(int w=0; w< wd/2;w++)	
					{
	
									// lower quarter image
						ResizeVLinePlanar(srcpp+pitch*((w*ht/wd)/2)+w,
										dstpp+pitch*(w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcpp+pitch*((w*ht/wd)/2)+(wd-w),
										dstpp+pitch*(w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
									// top quarter image
						ResizeVLinePlanar(srcpp+pitch*(ht-(w*ht/wd)/2)+w,
										dstpp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcpp+pitch*(ht-(w*ht/wd)/2)+(wd-w),
										dstpp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
					}				

					for(int h=0; h < ht/2; h++)
					{
										// Left quadrant image
						ResizeHLinePlanar(srcpp+pitch*h+((h*wd/ht)/2),
											dstpp+pitch*h+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcpp+pitch*(ht-h)+((h*wd/ht)/2),
											dstpp+pitch*(ht-h)+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
										// Right quadrant image
						ResizeHLinePlanar(srcpp+pitch*h+(wd-(h*wd/ht)/2),
											dstpp+pitch*h+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcpp+pitch*(ht-h)+(wd-(h*wd/ht)/2),
											dstpp+pitch*(ht-h)+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
					}
				}// int and nb > 8

				else if( fi->sampleType == stFloat)
				{

					const float * srcpp = (const float *)srcp;
					float *dstpp = (float *)dstp;

					for(int w=0; w< wd/2;w++)	
					{
	
									// lower quarter image
						ResizeVLinePlanar(srcpp+pitch*((w*ht/wd)/2)+w,
										dstpp+pitch*(w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcpp+pitch*((w*ht/wd)/2)+(wd-w),
										dstpp+pitch*(w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
									// top quarter image
						ResizeVLinePlanar(srcpp+pitch*(ht-(w*ht/wd)/2)+w,
										dstpp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+w,
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
						ResizeVLinePlanar(srcpp+pitch*(ht-(w*ht/wd)/2)+(wd-w),
										dstpp+pitch*(ht-w*(ht/2-2*d1y)/(wd))+(wd-w),
											w*(ht/2-2*d1y)/(wd/2),pitch,pitch,
											((float)(ht))/(ht-4*d1y));
					}				

					for(int h=0; h < ht/2; h++)
					{
										// Left quadrant image
						ResizeHLinePlanar(srcpp+pitch*h+((h*wd/ht)/2),
											dstpp+pitch*h+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcpp+pitch*(ht-h)+((h*wd/ht)/2),
											dstpp+pitch*(ht-h)+(h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
										// Right quadrant image
						ResizeHLinePlanar(srcpp+pitch*h+(wd-(h*wd/ht)/2),
											dstpp+pitch*h+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
						ResizeHLinePlanar(srcpp+pitch*(ht-h)+(wd-(h*wd/ht)/2),
											dstpp+pitch*(ht-h)+(wd-h*(wd/2-2*d1x)/(ht)),
											h*(wd/2-2*d1x)/(ht/2),
											((float)wd)/(wd-4*d1x));
					}
				}// float

			} // plane

		} // not crumple
	
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC crumpleFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    CrumpleData *d = (CrumpleData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC crumpleCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    CrumpleData d;
    CrumpleData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Crumple: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Crumple: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Crumple: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	
	temp = !! int64ToIntS(vsapi->propGetInt(in, "crumple", 0, &err));
	if(err)
		d.crumple = true;
	else if(temp == 0)
		d.crumple = false;
	else 
		d.crumple = true;

	temp = !! vsapi->propGetInt(in, "emerge", 0, &err);
	if(err)
		d.emerge = true;
	else if(temp == 0)
		d.emerge = false;
	else 
		d.emerge = true;

			 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (CrumpleData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Crumple", crumpleInit, crumpleGetFrame, crumpleFree, fmParallel, 0, data, core);
}

// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "central", "VapourSynth transition function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Crumple", "clip:clip;clipb:clip;overlap:float:opt;crumple:int:opt;emerge:int:opt;", crumpleCreate, 0, plugin);
}
*/