/******************************************************************************
Trans.Door filter function for vapoursynth by V.C.Mohan
The A Clip  unmasks a stationary B Clip clip opening  vertical or horizontal doors or
The B Clip masks  the A clip by closing vertical or horizontal doors.
 The clips should have identical lengths, height, width and same pixel type 
 5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	bool open; // clip B frames open or clip A frames inwards
	bool vert;	// veretical doors
	
	int vtransStart;
	int vtransEnd;
	int numFrames;
		
}DoorData;

//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC doorInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    DoorData *d = (DoorData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	
}
	
/***************************************************************/	
static const VSFrameRef *VS_CC doorGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     DoorData *d = (DoorData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);

		const VSFrameRef *base;	// stationary frame
		const VSFrameRef *src; // changing frame
		int deltay =((bn + 1) * height)/(2 * (d->overlap + 1));
		deltay = deltay & 0xFFFFFFFE;
		int deltax = ((bn + 1) * width)/(2 * (d->overlap + 1));
		deltax = deltax & 0xFFFFFFFE;
		int taperx = ((bn + 1) * (width /10) )/(2 * (d->overlap + 1));
		int tapery = ((bn + 1) * (height /10) )/(2 * (d->overlap + 1));

		if(deltax <= 2 || deltay <= 2)
		{
			vsapi->freeFrame(srcb);
			return srca;
		}

		int doorht, doorwd;

		if( d->vert)	// vertical doors			
		{
				// Vertical Doors open revealing Right frame
			
			doorht = height;
			doorwd = width / 2 - deltax;			
		}

		else // horizontal door opening
		{
			doorht = height /2 - deltay;
			doorwd = width;
		}

		if(d->open)
		{
			src = srca;
			base = srcb;
			
		}
		else
		{
			base = srca;
			src = srcb;
			if (d->vert)
			{
				
				tapery = height / 10 - tapery;
			}
			else
			{
					//  doors close
				
				taperx = width / 10 - taperx;
			}

		}

		float sq[3][3];
		float quad[4][2];
		float inv[3][3];

		int subh[] = {0, 0, 0};
		int subw[] = {0, 0, 0};

		if (fi ->colorFamily == cmYUV)
		{				
			subw[1] = fi->subSamplingW;
			subh[1] = fi->subSamplingH;
			subw[2] = fi->subSamplingW;
			subh[2] = fi->subSamplingH;
		}

		VSFrameRef * dst = vsapi->copyFrame(base, core);

		const uint8_t * srcp[] = {NULL, NULL, NULL};
		uint8_t * dstp[] = {NULL, NULL, NULL};
		int pitch[] = {0, 0, 0};

		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;
		
		for(int plane = 0; plane < fi->numPlanes; plane ++)
		{
			srcp[plane] = vsapi->getReadPtr(src, plane);
            pitch[plane] = vsapi->getStride(src, plane) / kb;
            dstp[plane] = vsapi->getWritePtr(dst, plane);
		}
		
						
			if(d->vert)
			{
				
				// rectangular src door original coord
				int xleft = 0;
				int xright = width / 2 - 1;
				int ytop = 0;
				int ybot = height - 1;
				// Quadrilateral  door changed coordinates
				int ltopx = 0;
				int lbotx = 0;
				int ltopy = 0;
				int lboty = height - 1;
				int rtopx = doorwd;
				int rtopy = tapery;
				int rbotx = rtopx;
				int rboty = height - 1 - tapery;

				// current quadrilateral coordinates
				quad[0][0] = (float)ltopx;
				quad[0][1] = (float)ltopy;
				quad[1][0] = (float)rtopx;
				quad[1][1] = (float)rtopy;
				quad[2][0] = (float)rbotx;
				quad[2][1] = (float)rboty;
				quad[3][0] = (float)lbotx;
				quad[3][1] = (float)lboty;

				// generate forward and inverse matrices 

				int transx = xleft;
				int transy = ytop;
				float scalex = (float)(xright - xleft); // / bwd;
				float scaley = (float)(ybot - ytop); // / bht;

				if( UnitSq2Quad( sq, inv, quad) != 0)
				{
					// error
					vsapi->freeFrame(base);
					return src;
				}

				int andh = (1 << subh[1]) - 1;
				int andw = (1 << subw[1]) - 1;

				for( int h = 0, hpitch = 0; h < height; h ++, hpitch += pitch[0])
				{
					for(int w = 0; w < rbotx; w ++)
					{
						float denom = inv[0][2]* w + inv[1][2] * h + inv[2][2];

						if( denom > -1e-10 && denom < 1e-10 ) continue;

						float sourcex = (inv[0][0] * w + inv[1][0] * h + inv[2][0]) / denom;

						int sx = (int)(sourcex * scalex + transx);

						float sourcey = (inv[0][1] * w + inv[1][1] * h + inv[2][1]) / denom;

						int sy = (int)(sourcey * scaley + transy);

						if( sx < xleft || sx > xright || sy < ytop || sy > ybot)
							continue;

						if ( fi->sampleType == stInteger && nb == 8)
						{
							uint8_t * dp = dstp[0];
							const uint8_t * sp = srcp[0];
							uint8_t * d1p = dstp[1];
							const uint8_t * s1p = srcp[1];
							uint8_t * d2p = dstp[2];
							const uint8_t * s2p = srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[hpitch + width - 1 - w]  = sp[ sy * pitch[0] + width - 1 - sx];

							if( ( h & andh) == 0 && (w & andw ) == 0 && fi->numPlanes > 1)
							{
								d1p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s1p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d1p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s1p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

								d2p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s2p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d2p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s2p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

							}
						}
						else if ( fi->sampleType == stInteger && nb > 8)
						{
							uint16_t * dp = (uint16_t *)dstp[0];
							const uint16_t * sp = (const uint16_t *) srcp[0];
							uint16_t * d1p = ( uint16_t *)dstp[1];
							const uint16_t *  s1p = (const uint16_t *)srcp[1];
							uint16_t * d2p = (uint16_t *)dstp[2];
							const uint16_t * s2p = (const uint16_t *)srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[hpitch + width - 1 - w]  = sp[ sy * pitch[0] + width - 1 - sx];

							if( ( h & andh) == 0 && (w & andw ) == 0  && fi->numPlanes > 1)
							{
								d1p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s1p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d1p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s1p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

								d2p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s2p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d2p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s2p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

							}
						}
						
						else if ( fi->sampleType == stFloat)
						{
							float* dp = (float*)dstp[0];
							const float* sp = (const float*) srcp[0];
							float* d1p = ( float*)dstp[1];
							const float*  s1p = (const float*)srcp[1];
							float* d2p = (float*)dstp[2];
							const float* s2p = (const float*)srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[hpitch + width - 1 - w]  = sp[ sy * pitch[0] + width - 1 - sx];

							if( ( h & andh) == 0 && (w & andw ) == 0  && fi->numPlanes > 1)
							{
								d1p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s1p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d1p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s1p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

								d2p[(h >> subh[1]) * pitch[1] + (w >> subw[1])]  
								= s2p[ (sy >> subh[1]) * pitch[1] + (sx >> subw[1] )];
								d2p[(h >> subh[1]) * pitch[1] + ((width - 1 - w) >> subw[1])]  
								= s2p[  (sy >> subh[1]) * pitch[1] + (( width - 1 - sx ) >> subw[1] )];

							}
						}
					}
				}
			}	// if d->vert

			else //	if(! d->vert)
			{				  
				// rectangular left door original coord
				int xleft = 0;
				int xright = width - 1;
				int ytop = 0;
				int ybot = height / 2 - 1;
				// Quadrilateral left door changed coordinates
				int ltopx = 0;
				int lbotx = taperx;
				int ltopy = 0;
				int lboty = doorht;
				int rtopx = width - 1 ;
				int rtopy = 0;
				int rbotx = width - 1 - taperx;
				int rboty = doorht ;

				// current quadrilateral coordinates
				quad[0][0] = (float)ltopx;
				quad[0][1] = (float)ltopy;
				quad[1][0] = (float)rtopx;
				quad[1][1] = (float)rtopy;
				quad[2][0] = (float)rbotx;
				quad[2][1] = (float)rboty;
				quad[3][0] = (float)lbotx;
				quad[3][1] = (float)lboty;

				// generate forward and inverse matrices 

				int transx = xleft;
				int transy = ytop;
				float scalex = (float)(xright - xleft); // / bwd;
				float scaley = (float)(ybot - ytop); // / bht;

				if( UnitSq2Quad( sq, inv, quad) != 0)
				{
					// error
					vsapi->freeFrame(base);
					return src;
				}

				int andh = (1 << subh[1]) - 1;
				int andw = (1 << subw[1]) - 1;

				int botpitch = (height - 1) * pitch[0];

				for( int h = 0, hpitch = 0; h < lboty; h ++, hpitch += pitch[0])
				{
					for(int w = 0; w < rtopx; w ++)
					{
						float denom = inv[0][2]* w + inv[1][2] * h + inv[2][2];

						if( denom > -1e-10 && denom < 1e-10 ) continue;

						float sourcex = (inv[0][0] * w + inv[1][0] * h + inv[2][0]) / denom;

						int sx = (int)(sourcex * scalex + transx);

						float sourcey = (inv[0][1] * w + inv[1][1] * h + inv[2][1]) / denom;

						int sy = (int)(sourcey * scaley + transy);

						if( sx < xleft || sx > xright || sy < ytop || sy > ybot)
							continue;

						if ( fi->sampleType == stInteger && nb == 8)
						{
							uint8_t * dp = dstp[0];
							const uint8_t * sp = srcp[0];
							uint8_t * d1p = dstp[1];
							const uint8_t * s1p = srcp[1];
							uint8_t * d2p = dstp[2];
							const uint8_t * s2p = srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[botpitch - hpitch + w]  = sp[ botpitch - sy * pitch[0] + sx];

							if( (h & andh) == 0 && (w & andw) == 0  && fi->numPlanes > 1)
							{
								d1p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s1p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d1p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s1p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];

								d2p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s2p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d2p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s2p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
							}

						}
						else if ( fi->sampleType == stInteger && nb > 8)
						{
							uint16_t * dp = (uint16_t *)dstp[0];
							const uint16_t * sp = (const uint16_t *) srcp[0];
							uint16_t * d1p = ( uint16_t *)dstp[1];
							const uint16_t *  s1p = (const uint16_t *)srcp[1];
							uint16_t * d2p = (uint16_t *)dstp[2];
							const uint16_t * s2p = (const uint16_t *)srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[botpitch - hpitch + w]  = sp[ botpitch - sy * pitch[0] + sx];

							if( (h & andh) == 0 && (w & andw) == 0  && fi->numPlanes > 1)
							{
								d1p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s1p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d1p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s1p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];

								d2p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s2p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d2p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s2p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
							}
						}
						else if ( fi->sampleType == stFloat)
						{
							float* dp = (float*)dstp[0];
							const float* sp = (const float*) srcp[0];
							float* d1p = ( float*)dstp[1];
							const float*  s1p = (const float*)srcp[1];
							float* d2p = (float*)dstp[2];
							const float* s2p = (const float*)srcp[2];

							dp[hpitch + w]  = sp[ sy * pitch[0] + sx];
							dp[botpitch - hpitch + w]  = sp[ botpitch - sy * pitch[0] + sx];

							if( (h & andh) == 0 && (w & andw) == 0  && fi->numPlanes > 1)
							{
								d1p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s1p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d1p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s1p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];

								d2p [(h >> subh[1]) * pitch[1] + (w >> subw[1] ) ]
								= s2p[ (sy >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
								d2p[ ((height - 1 - h) >> subh[1] ) *pitch[1] + ( w >> subw[1])]  
								= s2p[( (height - 1 - sy) >> subh[1] ) * pitch[1] + (sx >> subw[1]) ];
							}
						} 
					}
				}
			}	// if not d->vert

			// plane
		vsapi->freeFrame(src);
		return dst;
	}
	return 0;
}

//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC doorFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    DoorData *d = (DoorData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC doorCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    DoorData d;
    DoorData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Door: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Door: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Door: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	temp = !!vsapi->propGetInt(in, "open", 0, &err);
	if( err)
		d.open = true;

	if( temp == 0)
	{
		d.open = false;
	}
	else
		d.open = true;

	temp = !!vsapi->propGetInt(in, "vert", 0, &err);
	if( err)
		d.vert = true;

	else if( temp == 0)
	{
		d.vert = false;
	}
	else
		d.vert = true;
		

	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (DoorData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Door", doorInit, doorGetFrame, doorFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "door", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Door", "clip:clip;clipb:clip;overlap:float:opt;open:int:opt;vert:int:opt;", doorCreate, 0, plugin);
}
*/