/******************************************************************************
TransPaint filter plugin for Avisynth by V.C.Mohan
On The A Clip  parts of B clip are painted/built in the manner specified.
 The clips should have identical lengths, height, width and same pixel type 

5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
//------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int style;	// 1 fence, 2 paint, 3 rings 4

	int vtransStart;
	int vtransEnd;
	int numFrames;

	int * xyarr;
	
}PaintData;

//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC paintInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    PaintData *d = (PaintData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
	d->xyarr = NULL;
	// style 1 = fence, 2 = paint, 3 = rings, 4 = bricks
	if(d->style == 1 || d->style == 2 || d->style == 3 )
	{
	
		d->xyarr = (int*)vs_aligned_malloc<int>(sizeof(int) * (d->overlap + 3), 32); // new int[d->overlap + 3];
		int i = 0;
	
		// we need non repeating numbers
		while( i < d->overlap)
		{
			d->xyarr[i]= rand() % d->overlap;

			int j = 0;

			for(j = 0; j < i; j ++)

				if(d->xyarr[i] == d->xyarr[j])	// we got this number before
					break;
			if(j == i)	// we reached end . so increase it by one
			
				i ++;
		}
		// add 3 values to avoid mod in every access
		d->xyarr[d->overlap] = d->xyarr[0];
		d->xyarr[d->overlap + 1] = d->xyarr[1];
		d->xyarr[d->overlap + 2] = d->xyarr[2];
		
	}
}

/***************************************************************/	
static const VSFrameRef *VS_CC paintGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     PaintData *d = (PaintData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// make A frame as base on which to paint
		VSFrameRef * dst = vsapi->copyFrame(srca, core);

		
		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;
		if ( d->style == 1 || d->style == 2)
		{
			for ( int plane = 0; plane < fi->numPlanes; plane ++)
			{
				const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
				const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
				int stride = vsapi->getStride(srca, plane);
				uint8_t *dstp = vsapi->getWritePtr(dst, plane);
				int ht = vsapi->getFrameHeight(srca, plane);
				int wd = vsapi->getFrameWidth(srca, plane);
				int pitch = stride / nbytes;			
				int subw = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
				int subh = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingH : 0;

				if(d->style == 1)	// fence
				{
					int supportBoardWidth = height / 16;
					supportBoardWidth = (supportBoardWidth < 8) ? 8 : supportBoardWidth;
					supportBoardWidth >>= subh;

					// now  fence stakes width
					float brw = ((float) width) / d->overlap;
					if (brw < 4.0)
						brw = 4.0;
					brw /= (1 << subw );

					int dy=(8 * ht) / 9;	// stake height arbitrary
					float dx = (wd - brw) / d->overlap;	// stake slant width

					int mod4 = 4 >> subw, mod3 = 3 >> subw, mod2 = 2 >> subw;
					int mod10 = 10 >> subw;					

					if( fi->sampleType == stInteger && nbytes == 1)
					{
						uint8_t * dp = (uint8_t *) dstp;
						const uint8_t * sp = (const uint8_t *) srcpb;
						// build fence post supporting structure
						for (int w = 0; w < wd; w ++)
						{
							for(int h = 0; h < supportBoardWidth + (w % mod4); h ++)
							{
				
								dp[ (ht / 4 + h + (w % mod3) ) * pitch + w ] 
									=  sp[(ht / 4 + h + (w % mod3) ) * pitch + w];
								dp[ (3 * ht / 4 + h + (w % mod2) ) * pitch + w ] 
									=  sp[(3 * ht / 4 + h + (w % mod2) ) * pitch + w];
							}
						}
						
							// build fence stakes
						for(int stroke = 0; stroke < bn; stroke ++)
						{
							int sx = (int)(d->xyarr[stroke] * brw);
							int sy = d->xyarr[stroke+1] % 10;

							for( int h = sy; h < sy + dy; h ++)
							{
								int w = (int)(sx +((h - sy)*dx)/dy);
			
								for(int b = 0; b < brw; b ++)
								{				
									if( (h + (b % mod10) < ht - 1)  && (b + w) < wd-1) 
									{						
					
										dp[(h + (b % mod10)) * pitch + (b + w)]
										= sp[(h + (b % mod10)) * pitch + (b + w)];
									}
								}
							}
						}	// stroke					
					}	// stInteger nbits = 8

					if( fi->sampleType == stInteger && nbytes == 2)
					{
						uint16_t * dp = (uint16_t *) dstp;
						const uint16_t * sp = (const uint16_t *) srcpb;
						// build fence post supporting structure
						for (int w = 0; w < wd; w ++)
						{
							for(int h = 0; h < supportBoardWidth + (w % mod4); h ++)
							{
				
								dp[ (ht / 4 + h + (w % mod3) ) * pitch + w ] 
									=  sp[(ht / 4 + h + (w % mod3) ) * pitch + w];
								dp[ (3 * ht / 4 + h + (w % mod2) ) * pitch + w ] 
									=  sp[(3 * ht / 4 + h + (w % mod2) ) * pitch + w];
							}
						}
						
							// build fence stakes
						for(int stroke = 0; stroke < bn; stroke ++)
						{
							int sx = (int)(d->xyarr[stroke] * brw);
							int sy = d->xyarr[stroke+1] % 10;

							for( int h = sy; h < sy + dy; h ++)
							{
								int w = (int)(sx +((h - sy)*dx)/dy);
			
								for(int b = 0; b < brw; b ++)
								{				
									if( (h + (b % mod10) < ht - 1)  && (b + w) < wd-1) 
									{						
					
										dp[(h + (b % mod10)) * pitch + (b + w)]
										= sp[(h + (b % mod10)) * pitch + (b + w)];
									}
								}
							}
						}	// stroke					
					}	// stInteger nbits > 8

					if( fi->sampleType == stFloat)
					{
						float * dp = (float *) dstp;
						const float * sp = (const float *) srcpb;
						// build fence post supporting structure
						for (int w = 0; w < wd; w ++)
						{
							for(int h = 0; h < supportBoardWidth + (w % mod4); h ++)
							{
				
								dp[ (ht / 4 + h + (w % mod3) ) * pitch + w ] 
									=  sp[(ht / 4 + h + (w % mod3) ) * pitch + w];
								dp[ (3 * ht / 4 + h + (w % mod2) ) * pitch + w ] 
									=  sp[(3 * ht / 4 + h + (w % mod2) ) * pitch + w];
							}
						}
						
							// build fence stakes
						for(int stroke = 0; stroke < bn; stroke ++)
						{
							int sx = (int)(d->xyarr[stroke] * brw);
							int sy = d->xyarr[stroke+1] % 10;

							for( int h = sy; h < sy + dy; h ++)
							{
								int w = (int)(sx +((h - sy)*dx)/dy);
			
								for(int b = 0; b < brw; b ++)
								{				
									if( (h + (b % mod10) < ht - 1)  && (b + w) < wd-1) 
									{						
					
										dp[(h + (b % mod10)) * pitch + (b + w)]
										= sp[(h + (b % mod10)) * pitch + (b + w)];
									}
								}
							}
						}	// stroke					
					}	// stFloat

				}	// if style = 1 fence

				else if(d->style == 2)	// paint strokes	
				{
					int mod4 = 4 >> subw;
					float brw = (float)ht/(d->overlap);	// brush width		
								// paint brush starting x position		
					int dx = (wd / 16) & 0xFFFFFFFE;
	
					int dryup = wd / 10;	// paint brush dries up at end of stroke
				
					if( fi->sampleType == stInteger && nbits == 8)
					{
						uint8_t * dp = (uint8_t *) dstp;
						const uint8_t * sp = (const uint8_t *) srcpb;
				
						// left to right brush strokes drying up end
						for(int stroke = 0; stroke < bn + 1; stroke ++ )
						{
							int sx = ( (d->xyarr[stroke] * wd) / (10 * d->overlap)) & 0xFFFFFFFE;

							int sy = (int)(d->xyarr[stroke + 1] * brw);

							for(int b = 0; b < brw; b ++ )
							{
					
								for(int w = sx; w < sx + dx + dryup * (b % mod4) && w < wd; w ++ )
								{
									int h = (int)(sy + (w * 2 * brw) / dx);

									if((h + b) > 0 && (h + b) < ht)
									{
								
										dp[(h + b) * pitch + w] 
										= sp[(h + b) * pitch + w];

										dp[(h + b) * pitch + wd - w] 
										= sp[(h + b) * pitch + wd - w];
								
									}
								}
							}
						} // stroke
					} // stinteger, nbits = 8

					if( fi->sampleType == stInteger && nbits > 8)
					{
						uint16_t * dp = (uint16_t *) dstp;
						const uint16_t * sp = (const uint16_t *) srcpb;
				
						// left to right brush strokes drying up end
						for(int stroke = 0; stroke < bn + 1; stroke ++ )
						{
							int sx = ( (d->xyarr[stroke] * wd) / (10 * d->overlap)) & 0xFFFFFFFE;

							int sy = (int)(d->xyarr[stroke + 1] * brw);

							for(int b = 0; b < brw; b ++ )
							{
					
								for(int w = sx; w < sx + dx + dryup * (b % mod4) && w < wd; w ++ )
								{
									int h = (int)(sy + (w * 2 * brw) / dx);

									if((h + b) > 0 && (h + b) < ht)
									{
								
										dp[(h + b) * pitch + w] 
										= sp[(h + b) * pitch + w];

										dp[(h + b) * pitch + wd - w] 
										= sp[(h + b) * pitch + wd - w];
								
									}
								}
							}
						} // stroke
					} // stinteger, nbits > 8

					if( fi->sampleType == stFloat)
					{
						float * dp = (float *) dstp;
						const float * sp = (const float *) srcpb;
				
						// left to right brush strokes drying up end
						for(int stroke = 0; stroke < bn + 1; stroke ++ )
						{
							int sx = ( (d->xyarr[stroke] * wd) / (10 * d->overlap)) & 0xFFFFFFFE;

							int sy = (int)(d->xyarr[stroke + 1] * brw);

							for(int b = 0; b < brw; b ++ )
							{
					
								for(int w = sx; w < sx + dx + dryup * (b % mod4) && w < wd; w ++ )
								{
									int h = (int)(sy + (w * 2 * brw) / dx);

									if((h + b) > 0 && (h + b) < ht)
									{
								
										dp[(h + b) * pitch + w] 
										= sp[(h + b) * pitch + w];

										dp[(h + b) * pitch + wd - w] 
										= sp[(h + b) * pitch + wd - w];
								
									}
								}
							}
						} // stroke
					} // stFloat
				} // style = 2 for paint

			} // plane = 0..<3

		} // style 1 or 2

		else // styles 3 and 4
		{		
			const uint8_t * srcpb[] = {NULL,NULL,NULL};
			uint8_t * dstp[]  = {NULL,NULL,NULL};
			int pitch[3] ;
			int ht[3];
			int wd [3];
			int subh[3];
			int subw[3];

			for ( int plane = 0; plane < fi->numPlanes; plane ++)
			{
				const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
				srcpb[plane] = vsapi->getReadPtr(srcb, plane);
				int stride = vsapi->getStride(srca, plane);
				dstp[plane] = vsapi->getWritePtr(dst, plane);
				ht[plane] = vsapi->getFrameHeight(srca, plane);
				wd[plane] = vsapi->getFrameWidth(srca, plane);
				pitch[plane] = stride / nbytes;
				subh[plane] = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingH : 0;
				subw[plane] = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
			}

			
			if(d->style == 3)	// rings
			{
				int rt = (int)((2.0f * width)  / d->overlap);				// ring thickness

				for(int stroke = 0; stroke < bn+1; stroke ++)
				{
					int srad = (d->xyarr[stroke]* height) / (4 * d->overlap) + 2 * rt;
					int sx= (d->xyarr[stroke+1] * width) / d->overlap;
					int sy = (d->xyarr[stroke +2] * height) / d->overlap;
		
					if( fi->sampleType == stInteger && nbits == 8)
					{
						if( fi->numPlanes == 1)
						{
							vsDiscoPlane(srcpb[0],  dstp[0], srad, sx, sy, ht[0], wd[0],
										 sx, sy, ht[0], wd[0], pitch[0], pitch[0], 1.0, false, rt);
						}

						else
						{

							vsDiscoYUVPlanes(srcpb[0],  dstp[0], srad, sx, sy, ht[0], wd[0],
													sx, sy, ht[0], wd[0], pitch[0], pitch[0],
													srcpb[1], srcpb[2], pitch[1],  
													dstp[1],  dstp[2],pitch[1],
													subh[1],  subw[1],
													1.0, false, rt);
						}
					
					}	// stInteger nbits == 8

					else if( fi->sampleType == stInteger && nbits > 8)
					{
						uint16_t * dp = (uint16_t *) dstp[0];
						uint16_t * dp1 = (uint16_t *) dstp[1];
						uint16_t * dp2 = (uint16_t *) dstp[2];

						const uint16_t * sp = (const uint16_t *) srcpb[0];
						const uint16_t * sp1 = (const uint16_t *) srcpb[1];
						const uint16_t * sp2 = (const uint16_t *) srcpb[2];

						if( fi->numPlanes == 1)
						{
							vsDiscoPlane(sp,  dp, srad, sx, sy, ht[0], wd[0],
										 sx, sy, ht[0], wd[0], pitch[0], pitch[0], 1.0, false, rt);
						}

						else
						{

							vsDiscoYUVPlanes(sp,  dp, srad, sx, sy, ht[0], wd[0],
											sx, sy, ht[0], wd[0], pitch[0], pitch[0],
											sp1, sp2, pitch[1],	dp1,  dp2,pitch[1],
											subh[1],  subw[1], 1.0, false, rt);
						}
					
					}	// stInteger nbits >8

					if( fi->sampleType == stFloat)
					{
						float * dp = (float *) dstp[0];
						float * dp1 = (float *) dstp[1];
						float * dp2 = (float *) dstp[2];

						const float * sp = (const float *) srcpb[0];
						const float * sp1 = (const float *) srcpb[1];
						const float * sp2 = (const float *) srcpb[2];

						if( fi->numPlanes == 1)
						{
							vsDiscoPlane(sp,  dp, srad, sx, sy, ht[0], wd[0],
										 sx, sy, ht[0], wd[0], pitch[0], pitch[0], 1.0, false, rt);
						}

						else
						{

							vsDiscoYUVPlanes(sp,  dp, srad, sx, sy, ht[0], wd[0],
											sx, sy, ht[0], wd[0], pitch[0], pitch[0],
											sp1, sp2, pitch[1],	dp1,  dp2,pitch[1],
											subh[1],  subw[1], 1.0, false, rt);
						}
					
					}	// stFloat
				}	// stroke

			}	// style == 3

			else if (d->style == 4)	// brick wall
			{	
				int brw = 1 + wd[0] / d->overlap;// brick width
				int brh = ht[0] / d->overlap;	// brick height
		
				if ( fi->sampleType == stInteger && nbits == 8)
				{
					for(int nbw = 0; nbw < bn + 1; nbw ++)
						for(int nbh = 0; nbh < bn; nbh ++)								
							for(int bw = (nbh % 2) * brw / 2; bw < (nbw + nbh +1) * brw -1 + (nbh % 2) * brw / 2; bw += brw)
								for(int h = nbh * brh; h < (nbh + 1) * brh - 1; h ++)
									if(h < (n + 1) * brh )							
										for(int w = bw; w < bw + brw - 1; w ++)
											if(w < wd[0])
											{
												for(int p = 0; p < fi->numPlanes; p ++)
												{
													uint8_t * dp = dstp[p];
													const uint8_t *sp = srcpb[p]; 

													dp[ ((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])]
														=sp[((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])];
												}
											}
				}

				else if ( fi->sampleType == stInteger && nbits > 8)
				{
					for(int nbw = 0; nbw < bn + 1; nbw ++)
						for(int nbh = 0; nbh < bn; nbh ++)								
							for(int bw = (nbh % 2) * brw / 2; bw < (nbw + nbh +1) * brw -1 + (nbh % 2) * brw / 2; bw += brw)
								for(int h = nbh * brh; h < (nbh + 1) * brh - 1; h ++)
									if(h < (n + 1) * brh )							
										for(int w = bw; w < bw + brw - 1; w ++)
											if(w < wd[0])
											{
												for(int p = 0; p < fi->numPlanes; p ++)
												{
													uint16_t * dp = (uint16_t *)dstp[p];
													const uint16_t *sp = (uint16_t *)srcpb[p]; 

													dp[ ((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])]
														=sp[((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])];
												}
											}
				}	// stInteger nbits > 8

				else if ( fi->sampleType == stFloat)
				{
					for(int nbw = 0; nbw < bn + 1; nbw ++)
						for(int nbh = 0; nbh < bn; nbh ++)								
							for(int bw = (nbh % 2) * brw / 2; bw < (nbw + nbh +1) * brw -1 + (nbh % 2) * brw / 2; bw += brw)
								for(int h = nbh * brh; h < (nbh + 1) * brh - 1; h ++)
									if(h < (n + 1) * brh )							
										for(int w = bw; w < bw + brw - 1; w ++)
											if(w < wd[0])
											{
												for(int p = 0; p < fi->numPlanes; p ++)
												{
													float * dp = (float *)dstp[p];
													const float *sp = (float *)srcpb[p]; 

													dp[ ((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])]
														=sp[((height - ( 1 + bn) * brh + h) >> subh[p]) * pitch[p] + (w >> subw[p])];
												}
											}
				}	// stFloat

			}	// styl == 4	

		}	// style == 3 || style == 4
										
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC paintFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    PaintData *d = (PaintData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	if (d->xyarr != NULL)
		vs_aligned_free(d->xyarr);
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC paintCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    PaintData d;
    PaintData *data;
    int err;
//	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Paint: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Paint: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Paint: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.style = int64ToIntS(vsapi->propGetInt(in, "style", 0, &err));
	if( err)
		d.style = 1;

	if( d.style < 1 || d.style > 4)
	{
		vsapi->setError(out, "Paint: valid values for style are 1 for fence, 2 for paint, 3 for rings and 4 for bricks ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (PaintData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Paint", paintInit, paintGetFrame, paintFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "paint", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Paint", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;", paintCreate, 0, plugin);
}
*/