/******************************************************************************
Plugin function for vapoursynth
The Trans.Ripple Filter creates a ripple over which transition occurs.
During first 1/3 of transition ripple grows on A clip,
next 1/3 from within ripple the B clip ripple emerges 
 and during rest the ripple subsides 

5 Aug 2015  3 june 2021, 17 july 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int lambda;		// wave length : 16 to 1/8 frame diagonal 
	int Amp;		// wave height : 2 to 12		
	int origin;	// 0:center, 1:north 2: ne, 3: east, 4:se, 5:south, 6:sw, 7: west 8: nw
	bool merge;	// use color merge to grey
	int vtransStart;
	int vtransEnd;
	int numFrames;		
	double speed;			// .0 to 100 speed of spreading of wave motion. 
	int radmax;		// Max radius ripple to spread by end of clip
	int xcoord, ycoord;	// origin
	float *sintbl;
}RippleData;


//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC rippleInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    RippleData *d = (RippleData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	int wd = d->vi.width;

	int ht = d->vi.height;

	// calculate x and y coordinates of centers of ripples and max radius of ripple
	
	if (d->origin == 0)	// center of frame
	{
		d->xcoord = wd/2;

		d->ycoord = ht / 2;

		d->radmax = (int)sqrt((float)(ht * ht + wd * wd)) / 2;
	}

	else if (d->origin == 4) // se corner of frame
	{
		d->radmax = (int)sqrt((float)(ht * ht + wd * wd));

		d->xcoord = wd -1;

		d->ycoord = ht - 1;
	}

	else if (d->origin == 6)	// sw corner
	{
		d->radmax = (int)sqrt((float)(ht * ht + wd * wd));

		d->xcoord = 0;

		d->ycoord = ht - 1;
	}

	else if (d->origin == 2)	// ne corner
	{
		d->radmax = (int)sqrt((float)(ht * ht + wd * wd));

		d->xcoord = wd - 1;

		d->ycoord = 0;
	}

	else if ( d->origin == 8)	// nw corner
	{
		d->radmax = (int)sqrt((float)(ht * ht + wd * wd));

		d->xcoord = 0;

		d->ycoord = 0;
	}

	else if (d->origin == 1)	// north
	{

		d->radmax = (int)sqrt((float)(ht * ht + (wd * wd)/ 4));

		d->xcoord = wd/2;

		d->ycoord = 0;
	}

	else if (d->origin == 5)	// south
	{
		d->radmax = (int)sqrt((float)(ht * ht + (wd * wd)/ 4));

		d->xcoord = wd/2;

		d->ycoord = ht - 1;
	}
	else if (d->origin == 3)	// east
	{
		d->radmax = (int)sqrt((float)((ht * ht) / 4 + (wd * wd)));

		d->xcoord = wd - 1;

		d->ycoord = ht / 2;
	}

	else if (d->origin == 7)	// west
	{
		d->radmax = (int)sqrt((float)((ht * ht) / 4 + (wd * wd)));

		d->xcoord = 0;

		d->ycoord = ht / 2;
	}
/*
	if ( d->vi.format->colorFamily == cmRGB )
	{
		d->ycoord = ht - 1 - d->ycoord;			
	}
*/
	d->speed = d->lambda / 4; 	// wave propogation speed
		
	d->sintbl = (float*)vs_aligned_malloc<float>(sizeof(float) * d->lambda, 32); // new float[d->lambda];

		// create sine table	this is the wave amplitude at any position along its wavelength
	for(int i = 0; i < d->lambda; i ++)

		d->sintbl[i] = (float)sin(i * M_PI / (d->lambda / 2));

}

/***************************************************************/	
static const VSFrameRef *VS_CC rippleGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     RippleData *d = (RippleData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;	

		int phase1 = (d->overlap )/ 3;  // A frame ripple starts and expands full size 0 to phase1
		int phase2 = d->overlap - phase1;	// over A frame ripple, B frame ripple starts and fully covers phase1 to phase2
											// last phase phase2 to overlap
		VSFrameRef * dst = (bn < phase2 ) ? vsapi->copyFrame ( srca, core) : vsapi->copyFrame(srcb, core);

		const uint8_t * srpa[] = {NULL, NULL, NULL};
		const uint8_t * srpb[] = {NULL, NULL, NULL};
		uint8_t * dsp[] = {NULL, NULL, NULL};
		int pitches[] = {0, 0, 0};;

		int subH = (fi->colorFamily == cmYUV) ? fi->subSamplingH : 0;
		int subW = (fi->colorFamily == cmYUV) ? fi->subSamplingW : 0;
		
		// as all plane pointers are needed for the ripple we get them here
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *  dstp = vsapi->getWritePtr(dst, plane);
		//	int ht = vsapi->getFrameHeight(srca, plane);
		//	int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / nbytes;	
			srpa[plane] = srcpa;
			srpb[plane] = srcpb;
			dsp[plane] = dstp;
			pitches[plane] = pitch;
		}

		if( bn <= phase1)
		{			
			int nn = (int)(bn * d->speed);
					// increase ripple radius  from  zero to max and grey for merge option in this phase
			int erad = (bn * d->radmax) / phase1;
			int irad = 0;
			int amp = d->Amp;
			// on dst we have srca frame
			if (fi->sampleType == stInteger && nbytes == 1)
			{
				uint8_t grey = 1 << (nbits - 1);
				uint8_t *dp = dsp[0];
				const uint8_t * sp = srpa[0];

				if ( fi->numPlanes == 1) 
				{
					// single plane ripple increase radius  to maximum by end of this ophase
					CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
				}

				else // YUV RGB
				{
					uint8_t *dpU = dsp[1];
					const uint8_t * spU = srpa[1];
					uint8_t *dpV = dsp[2];
					const uint8_t * spV = srpa[2];
										
					if( d->merge)
					{
						if  (fi->colorFamily == cmYUV)
							// YUV planes grey ripple increase radius  to maximum by end of this ophase
							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1- bn, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);

						else if  (fi->colorFamily == cmRGB)
							// RGB planes correspond to YUV planes pointers
							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1 - bn,  
												width, height,  amp,  d->lambda, d->sintbl);
					}
					else
						// 3 planes color ripple increase radius  to maximum by end of this ophase
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);					
				}					
			}

			else if (fi->sampleType == stInteger && nbytes == 2)
			{
				uint16_t grey = 1 << (nbits - 1);
				uint16_t *dp =  (uint16_t *)dsp[0];
				const uint16_t * sp = ( const uint16_t *)srpa[0];

				if ( fi->numPlanes == 1) 
				{
					
					// y8 format
					CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
				}

				else // YUV RGB
				{
					uint16_t *dpU = (uint16_t *)dsp[1];
					const uint16_t * spU = (const uint16_t *)srpa[1];
					uint16_t *dpV = (uint16_t *)dsp[2];
					const uint16_t * spV = (const uint16_t *)srpa[2];
					
					if( d->merge)
					{
						if  (fi->colorFamily == cmYUV)

							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1- bn, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);
						else if (fi->sampleType == cmRGB)

							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1 - bn,  
												width, height,  amp,  d->lambda, d->sintbl);
					}
					else
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
					
				}					
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = (float)0;
				float *dp =  (float *)dsp[0];
				const float * sp = ( const float *)srpa[0];

				if ( fi->numPlanes == 1) 
				{
					
					CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
				}

				else // YUV or rgb
				{
					float *dpU = (float *)dsp[1];
					const float * spU = (const float *)srpa[1];
					float *dpV = (float *)dsp[2];
					const float * spV = (const float *)srpa[2];
					
					if( d->merge)
					{
						if  (fi->colorFamily == cmYUV)

							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1- bn, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);

						else if (fi->sampleType == cmRGB)

							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, bn , phase1 - bn,  
												width, height,  amp,  d->lambda, d->sintbl);
					}
					else
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);					
				}					
			}
		}

		else if (bn <= phase2)
		{
			int nn = (int)(bn * d->speed);
					// in this phase right frame ripple starts and expands
					// keep end radius and amp at their full values in this phase
			int erad = d->radmax;
					// irad is start of left frame or end of right frame ripple
			int irad = ((bn - phase1) * d->radmax) / (phase2 - phase1); 
			int amp = d->Amp;	// full value
										
			if (fi->sampleType == stInteger && nbytes == 1)
			{
				uint8_t grey = 1 << (nbits - 1);
				uint8_t *dp = dsp[0];
				const uint8_t * spa = srpa[0];
				const uint8_t * spb = srpb[0];

				if ( fi->numPlanes == 1) 
				{
					// y8 format
					if ( d->merge)
					{
						// A frame y plane ripple fades and B frame y plane ripple fades out
							MergeGrayRipple( dp, pitches[0],spa, pitches[0], spb, pitches[0],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
					}

					else
					{
					
						// A frame ripple visible part
						CreateRipple(dp, pitches[0], spa, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
						// B Frame ripple expanding over A ripple
						CreateRipple(dp, pitches[0], spb, pitches[0],
								d->xcoord, d->ycoord, 0, irad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or  RGB
				{
					uint8_t *dpU = dsp[1];					
					uint8_t *dpV = dsp[2];
					const uint8_t * spaU = srpa[1];
					const uint8_t * spaV = srpa[2];
					const uint8_t * spbU = srpb[1];					
					const uint8_t * spbV = srpb[2];
					
					if( d->merge)
					{
						if ( fi->colorFamily == cmYUV)
						{
							// A frame y plane ripple fades and B frame y plane ripple fades out
							MergeGrayRipple( dp, pitches[0],spa, pitches[0], spb, pitches[0],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
							// grey out U and V planes
							ColorPlane(dpU, pitches[1],width >> subW, height >> subH, grey );
							ColorPlane(dpV, pitches[2],width >> subW, height >> subH, grey );
							
						}

						else if (fi->colorFamily == cmRGB)
						{							
								// Green is taken as Y an approximation
							MergeGrayRipple( dpU, pitches[1], spaU, pitches[1], spbU, pitches[1],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
							// green value copied for red and blue planes
							vs_bitblt(dsp[0], pitches[0] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );
							vs_bitblt(dsp[2], pitches[2] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );

						}
					}
					else	// no merging colors
					{
						// B ripple expands covering A ripple
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spb, pitches[0], spbU, spbV, pitches[1],
										d->xcoord, d->ycoord, 0,  irad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spa, pitches[0], spaU, spaV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
					}					
				}					
			}

			else if (fi->sampleType == stInteger && nbytes == 2)
			{
				uint16_t grey = 1 << (nbits - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * spa = (const uint16_t *)srpa[0];
				const uint16_t * spb = (const uint16_t *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
					// y8 format
					if ( d->merge)
					{
						// A frame y plane ripple fades and B frame y plane ripple fades out
							MergeGrayRipple( dp, pitches[0],spa, pitches[0], spb, pitches[0],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
					}

					else
					{
					
						// A frame ripple visible part
						CreateRipple(dp, pitches[0], spa, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
						// B Frame ripple expanding over A ripple
						CreateRipple(dp, pitches[0], spb, pitches[0],
								d->xcoord, d->ycoord, 0, irad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or  RGB
				{
					uint16_t *dpU = (uint16_t *)dsp[1];					
					uint16_t *dpV = (uint16_t *)dsp[2];
					const uint16_t * spaU = (const uint16_t *)srpa[1];
					const uint16_t * spaV = (const uint16_t *)srpa[2];
					const uint16_t * spbU = (const uint16_t *)srpb[1];					
					const uint16_t * spbV = (const uint16_t *)srpb[2];
					
					if( d->merge)
					{
						if ( fi->colorFamily == cmYUV)
						{
							// this ripple already covering full frame and grey

							MergeGrayRipple( dp, pitches[0],spa, pitches[0], spb, pitches[0],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
							// grey out U and V planes
							ColorPlane(dpU, pitches[1],width >> subW, height >> subH, grey );
							ColorPlane(dpV, pitches[2],width >> subW, height >> subH, grey );
							
						}

						else if (fi->colorFamily == cmRGB)
						{
							
								// Green is taken as Y
							MergeGrayRipple( dpU, pitches[1], spaU, pitches[1], spbU, pitches[1],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
							// grey out U and V planes
							vs_bitblt(dsp[0], pitches[0] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );
							vs_bitblt(dsp[2], pitches[2] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );

						}
					}
					else	// no merging colors
					{

						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spb, pitches[0], spbU, spbV, pitches[1],
										d->xcoord, d->ycoord, 0,  irad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spa, pitches[0], spaU, spaV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
					}					
				}					
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * spa = (const float *)srpa[0];
				const float * spb = (const float *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
					// y8 format
					if ( d->merge)
					{
						// A frame y plane ripple fades and B frame y plane ripple fades out
							MergeGrayRipple( dp, pitches[0],spa, pitches[0], spb, pitches[0],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
					}

					else
					{
					
						// A frame ripple visible part
						CreateRipple(dp, pitches[0], spa, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
						// B Frame ripple expanding over A ripple
						CreateRipple(dp, pitches[0], spb, pitches[0],
								d->xcoord, d->ycoord, 0, irad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or  RGB
				{
					float *dpU = (float *)dsp[1];					
					float *dpV = (float *)dsp[2];
					const float * spaU = (const float *)srpa[1];
					const float * spaV = (const float *)srpa[2];
					const float * spbU = (const float *)srpb[1];					
					const float * spbV = (const float *)srpb[2];
					
					if( d->merge)
					{
						if ( fi->colorFamily == cmYUV)
						{
							// this ripple already covering full frame and grey

							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spa, pitches[0], spaU, spaV, pitches[1],
										d->xcoord, d->ycoord, irad, erad, nn, 1, 0, grey, 
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
							// this ripple expands and gradually greys
							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spb, pitches[0], spbU, spbV, pitches[1],
										d->xcoord, d->ycoord, 0,  irad,  nn, 1 , 0, grey, 
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
						}

						else if  (fi->colorFamily == cmRGB) // RGB
						{

								// Green is taken as Y
							MergeGrayRipple( dpU, pitches[1], spaU, pitches[1], spbU, pitches[1],
											d->xcoord, d->ycoord, 0, d->radmax, nn,  phase2 - bn, bn - phase1,  
											width, height, amp,  d->lambda, d->sintbl);
							// green value for red and blue planes to make grey
							vs_bitblt(dsp[0], pitches[0] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );
							vs_bitblt(dsp[2], pitches[2] * nbytes, dsp[1], pitches[1] * nbytes, width * nbytes, height );
						}
					}
					else	// no merging colors
					{

						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spb, pitches[0], spbU, spbV, pitches[1],
										d->xcoord, d->ycoord, 0,  irad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										spa, pitches[0], spaU, spaV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);
					}
						
					
				}					
			}
		}

		else if (bn > phase2)	// last phase
		{
			// in this phase the b frame ripple which completely covered frame subsides and gets color		
			int erad = ( (d->overlap - bn ) * d->radmax) / (d->overlap - phase2);
			int irad = 0;
			int amp = d->Amp; // * (d->overlap - bn)) / (d->overlap - phase2);
			int nn = (int)(bn * d->speed);

			if (fi->sampleType == stInteger && nbytes == 1)
			{
				uint8_t grey = 1 << (nbits - 1);
				uint8_t *dp = dsp[0];
				const uint8_t * sp = srpb[0];

				if ( fi->numPlanes == 1) 
				{
					if (d->merge)
					{
						// ripple subsides revealing B frame
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, 0, d->radmax, nn, width, height,
								(amp * (d->overlap - bn)) / (d->overlap - phase2), d->lambda, d->sintbl);
					}

					else
					{
				
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or RGB
				{
					uint8_t *dpU = dsp[1];
					const uint8_t * spU = srpb[1];
					uint8_t *dpV = dsp[2];
					const uint8_t * spV = srpb[2];
					int amp = ((d->overlap - bn ) * d->Amp) / (d->overlap - phase2);
					uint8_t grey = 1 << (nbits - 1);
					// recover color, subside ripple
					
					if (d->merge)
					{
						if (fi->colorFamily == cmYUV)
						{

							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, 0, d->radmax, nn, 
												d->overlap - bn, bn - phase2, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);
						}
						else if (fi->colorFamily == cmRGB)	// RGB
						{

							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, 0, d->radmax, nn, d->overlap - bn, bn - phase2,  
												width, height, amp,  d->lambda, d->sintbl);
						}
					}
					

					else
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);					
				}
			}

			if (fi->sampleType == stInteger && nbytes == 2)
			{
				uint16_t grey = 1 << (nbits - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * sp = (const uint16_t *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
				
					if (d->merge)
					{
						// ripple subsides revealing B frame
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, 0, d->radmax, nn, width, height,
								(amp * (d->overlap - bn)) / (d->overlap - phase2), d->lambda, d->sintbl);
					}

					else
					{
				
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or RGB
				{
					uint16_t *dpU = (uint16_t *)dsp[1];
					const uint16_t * spU = (const uint16_t *)srpb[1];
					uint16_t *dpV = (uint16_t *)dsp[2];
					const uint16_t * spV = (const uint16_t *)srpb[2];
					//  
					if (d->merge)
					{
						if (fi->colorFamily == cmYUV)
						{
							// occupies full frame, reduces amplitude and returns to color
							ColorGrayRipple (dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, 0, d->radmax, nn, 
												d->overlap - bn, bn - phase2, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);
						}
						else if (fi->colorFamily == cmRGB)	// RGB
						{

							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, 0, d->radmax, nn, d->overlap - bn, bn - phase2,  
												width, height, amp,  d->lambda, d->sintbl);
						}
					}
					

					else
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);

					
				}
			}

			if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * sp = (const float *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
				
					if (d->merge)
					{
						// ripple subsides revealing B frame
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, 0, d->radmax, nn, width, height,
								(amp * (d->overlap - bn)) / (d->overlap - phase2), d->lambda, d->sintbl);
					}

					else
					{
				
						CreateRipple(dp, pitches[0], sp, pitches[0],
								d->xcoord, d->ycoord, irad, erad, nn, width, height,
								amp, d->lambda, d->sintbl);
					}
				}

				else // YUV or RGB
				{
					float *dpU = (float *)dsp[1];
					const float * spU = (const float *)srpb[1];
					float *dpV = (float *)dsp[2];
					const float * spV = (const float *)srpb[2];
					//  reduces  radius
					if (d->merge)
					{
						if (fi->colorFamily == cmYUV)
						{

							ColorGrayRipple ( dp, pitches[0], dpU, dpV, pitches[1],
												sp, pitches[0], spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, d->overlap - bn, bn - phase2, grey, 
												width, height, subH, subW, amp,  d->lambda, d->sintbl);
						}

						else if (fi->colorFamily == cmRGB)	// RGB
						{

							ColorGrayRipple ( dp,  dpU, dpV, pitches[1],
												sp,  spU, spV, pitches[1],
												d->xcoord, d->ycoord, irad, erad, nn, d->overlap - bn, bn - phase2,  
												width, height, amp,  d->lambda, d->sintbl);
						}
					}
					

					else
						CreateRipple ( dp, pitches[0], dpU, dpV, pitches[1],
										sp, pitches[0], spU, spV, pitches[1],
										d->xcoord, d->ycoord, irad,erad,  nn,
										width, height, subH, subW, amp,  d->lambda, d->sintbl);					
				}
			}
		}		
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC rippleFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    RippleData *d = (RippleData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	vs_aligned_free(d->sintbl);
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC rippleCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    RippleData d;
    RippleData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Ripple: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Ripple: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Ripple: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.origin = int64ToIntS(vsapi->propGetInt(in, "origin", 0, &err));
	if( err)
		d.origin = 0;

	if( d.origin < 0 || d.origin > 8)
	{
		vsapi->setError(out, "Ripple: valid values for origin are 0 for center and 1 to 8 for north to nw clockwise ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	d.Amp = int64ToIntS(vsapi->propGetInt(in, "amp", 0, &err));
	if( err)
		d.Amp = 4;	

	if( d.Amp < 1 || d.Amp > 12)
	{
		vsapi->setError(out, "Ripple: valid values for wave  amp are 1 to 12  ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	d.Amp *= 4;	// 1 is very low
	d.lambda = int64ToIntS(vsapi->propGetInt(in, "wlength", 0, &err));
	if( err)
		d.lambda = 40;

	if( d.lambda < 4 || d.lambda * d.lambda > d.vi.height * d.vi.height + d.vi.width * d.vi.width)
	{
		vsapi->setError(out, "Ripple: valid values for wlength are 4 to 1/8 frame diagonal ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	temp =  !! vsapi->propGetInt(in, "merge", 0, &err);
	if( err )
		d.merge = false;
	else if (temp == 1)
		d.merge = true;
	else
		d.merge = true;
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (RippleData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Ripple", rippleInit, rippleGetFrame, rippleFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "ripple", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Ripple", "clip:clip;clipb:clip;overlap:float:opt;origin:int:opt;amp:int:opt;lambda:int:opt;merge:int:opt;", rippleCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
