/******************************************************************************
Trans.Roll  plugin function for vapoursynth by V.C.Mohan
The A Clip  rolling unmasks a stationary B Clip clip in the direction specified.
 The clips should have identical lengths, height, width and same pixel type 
 Roll Diameter and
 shading along roll length are other params
 
 5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
********************************************************************************/  
//------------------------------------------------------------
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int maxroll;
	int dir;
	int shade;
	bool rollin;

	int vtransStart;
	int vtransEnd;
	int numFrames;
	
}RollData;
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC rollInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    RollData *d = (RollData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
	d->maxroll = 128;
}
	
/***************************************************************/	
static const VSFrameRef *VS_CC rollGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     RollData *d = (RollData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;//				 
		
		int deltay =(((bn+1) * height)/(d->overlap + 1)) & 0xFFFFFFF8;

		int deltax = (((bn+1) * width)/(d->overlap + 1)) & 0xFFFFFFF8;		

		int maxroll = d->maxroll;
		int  wht, wwd;	// current width of roll (dia) along y or x axis

		if( d->dir  == 1 || d->dir == 5 )
		{
			if(deltay < 8 )
			{
				vsapi->freeFrame(srcb);
				return srca;
			}

			wht = deltay / 2;	// current width of roll (dia) along y axis

			if(deltay > maxroll)

				wht = maxroll / 2;

			if(height - deltay < maxroll)

				wht = (height - deltay) / 2;

			if(wht < 4)
			{
				vsapi->freeFrame(srca);

				return srcb;
			}
		}

		if( d->dir  == 3 || d->dir == 7  )
		{
			if(deltax < 8 )
			{
				vsapi->freeFrame(srcb);
				return srca;
			}

			wwd = deltax/2;	// width of roll (dia) along y axis

			if(deltax > maxroll)
				wwd = maxroll / 2;

			if(width - deltax < maxroll)
				wwd = (width - deltax) / 2;

			if(wwd < 4)
			{
				vsapi->freeFrame(srca);

				return srcb;
			}
		}

		VSFrameRef *dst;
		const VSFrameRef * src;
		
		if( ((d->dir & 1) == 1 && ! d->rollin) || ((d->dir & 1) == 0 &&  d->rollin) )	
		{
			dst = vsapi->copyFrame ( srca, core);
			src = srcb;
		}

		else
		{
			dst = vsapi->copyFrame ( srcb, core);
			src = srca;
		}
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcp = vsapi->getReadPtr(src, plane);
			
			int stride = vsapi->getStride(src, plane);
			uint8_t *  dstp = vsapi->getWritePtr(dst, plane);
			int planeht = vsapi->getFrameHeight(srca, plane);
			int planewd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / kb;
			int subH = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingH : 0;
			int subW = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
			int andH = (1 << subH) -1;
			int andW = (1 << subW) -1;
			int dx = ((bn+1)* planewd)/(d->overlap+1) ;
			int dy = ((bn+1)* planeht)/(d->overlap+1) ;
			int wt = wht >> subH;			
			int wdd = wwd >> subW;
			if( ! d->rollin)
			{
				dx = planewd  - dx;
				dy = planeht - dy;
			}	

			if( ( d->rollin && d->dir == 5 ) || (!d->rollin && d->dir == 1) )	// down			
			{			// A Frame rolls downward unmasking RightFrame
					// copy undisturbed part of frame A
				vs_bitblt(dstp  + (dy + wt) * stride, stride, srcp  + (dy + wt) * stride, stride, planewd * kb, planeht  - dy - wt);

				int offsets = (dy - wt)  * pitch;
				int offsetd = (dy + wt / 2)  * pitch;

				if(fi->sampleType == stInteger && nb == 8)
				{
					uint8_t * dp  = dstp;					
					const uint8_t * sp  = srcp;
					uint8_t shade = 0;
					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}				
				}

				else if(fi->sampleType == stInteger && nb > 8)
				{
					uint16_t * dp  = (uint16_t *) dstp;					
					const uint16_t * sp  = (const uint16_t *)srcp;
					uint16_t shade = 0;

					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}				
				}

				else if(fi->sampleType == stFloat)
				{
					float * dp  = (float *) dstp;					
					const float * sp  = (const float *)srcp;
					float shade = 0;

					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}				
				}

			}
			

			else if( ( d->rollin && d->dir == 1 ) || (!d->rollin && d->dir == 5) )	// up		
			{
				// copy undisturbed part of frame A
				vs_bitblt( dstp, stride, srcp, stride, planewd * kb, planeht - dy - wt);

				int offsets = ( planeht - 1 -dy + wt)  * pitch;
				int offsetd = (planeht - 1 - dy - wt / 2)  * pitch;

						// A Frame rolls upward unmasking B Frame			
				if(fi->sampleType == stInteger && nb == 8)
				{
					uint8_t * dp  = dstp;					
					const uint8_t * sp  = srcp;
					uint8_t shade = 0;

					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}			
				}

				else if(fi->sampleType == stInteger && nb > 8)
				{
					uint16_t * dp  = (uint16_t *) dstp;					
					const uint16_t * sp  = (const uint16_t *)srcp;
					uint16_t shade = 0;

					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}			
				}

				else if(fi->sampleType == stFloat)
				{
					float * dp  = (float *) dstp;					
					const float * sp  = (const float *)srcp;
					float shade = 0;

					for(int w = 0; w < planewd; w ++)
					{
						RollzVLinePlanar(sp + offsets + w,
											dp + offsetd + w,
											wt , pitch, pitch, shade);
					}			
				}
			}

			
			else if( ( d->rollin && d->dir == 3 ) || (!d->rollin && d->dir == 7) )		// towards right	
			{		//A Frame  rolls out  Right unmasking B Frame or B rollsin
				// copy undisturbed part of frame A
				vs_bitblt(dstp + (dx + wdd)* kb, stride, srcp + (dx + wdd ) * kb , stride, (planewd - dx -wdd) * kb, planeht);

				int offsets = (dx - wdd) ;
				int offsetd = (dx + wdd / 2) ;

				if(fi->sampleType == stInteger && nb == 8)
				{
					uint8_t * dp  = dstp;					
					const uint8_t * sp  = srcp;
					uint8_t shade = 0;					

					for(int h = 0; h <  planeht; h ++)				
					{
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}			
				}

				else if(fi->sampleType == stInteger && nb > 8)
				{
					uint16_t * dp  = (uint16_t *) dstp;					
					const uint16_t * sp  = (const uint16_t *)srcp;
					uint16_t shade = 0;					

					for(int h = 0; h <  planeht; h ++)				
					{
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}			
				}

				else if(fi->sampleType == stFloat)
				{
					float * dp  = (float *) dstp;					
					const float * sp  = (const float *)srcp;
					float shade = 0;				

					for(int h = 0; h <  planeht; h ++)				
					{
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}			
				}
			}			

			else if( ( d->rollin && d->dir == 7 ) || (!d->rollin && d->dir == 3) )			
			{			//A Frame rolls westward unmasking B Frame
				// copy undisturbed part of frame A
				vs_bitblt(dstp, stride, srcp, stride, (planewd - dx -wdd) * kb, planeht);

				int offsets = (planewd - 1 - dx + wdd);
				int offsetd = (planewd - 1 - dx - wdd / 2) ;

				if(fi->sampleType == stInteger && nb == 8)
				{
					uint8_t * dp  = dstp;					
					const uint8_t * sp  = srcp;
					uint8_t shade = 0;

					for(int h = 0; h <  planeht; h ++)				
					{
					
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}				
				}

				else if(fi->sampleType == stInteger && nb > 8)
				{
					uint16_t * dp  = (uint16_t *) dstp;					
					const uint16_t * sp  = (const uint16_t *)srcp;
					uint16_t shade = 0;

					for(int h = 0; h <  planeht; h ++)				
					{
					
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}				
				}

				else if(fi->sampleType == stFloat)
				{
					float * dp  = (float *) dstp;					
					const float * sp  = (const float *)srcp;
					float shade = 0;

					for(int h = 0; h <  planeht; h ++)				
					{
					
						RollzHLinePlanar(sp + offsets,
										dp + offsetd,
										wdd, shade);

						offsets += pitch;
						offsetd += pitch;
					}				
				}
			}	// dir

			else if( (d->dir & 1 ) == 0)
			{
				int dx = 2 *((bn+1)* planewd)/(d->overlap+1) ;
			int dy = 2 *((bn+1)* planeht)/(d->overlap+1) ;

			if (  dx < 2 || dy < 2)
			{
				vsapi->freeFrame (srcb);
				return srca;
			}

			if( ! d->rollin)
			{
				dx = 2 * planewd  - dx;
				dy = 2 * planeht - dy;
			}					

			if( ( d->rollin && d->dir ==6)  || (!d->rollin && d->dir == 2) )// "sw")
			{
				for ( int h = 0; h < dy; h ++)
				{
					int ww = ((dy - h) * planewd) / planeht;
					ww = ww > planewd -1? planewd -1 : ww < 0? 0: ww;

					int hh =  h;
					if ( hh > planeht - 2 || hh < 2) continue;

					vs_bitblt(dstp + hh * stride + (planewd - 1 - ww) * kb, stride, 
								srcp + hh * stride + (planewd - 1 -ww) * kb, stride, ww * kb, 1);
					int rolldia = 0;
					if(  ww < maxroll)
						rolldia = ww;
					else if( ww > planewd - maxroll)
						rolldia = planewd - ww ;
					else
						rolldia = maxroll;
					if(rolldia > ( 8 >> subW) )
					{
						if(fi->sampleType == stInteger && nb == 8)
						{	
							uint8_t shade = 0;
						
							RollzHLinePlanar( srcp + hh * pitch + (planewd - 1 - ww),  
												dstp + hh * pitch + (planewd - 1 - ww), rolldia/2, shade);
						}

						else if(fi->sampleType == stInteger && nb > 8)
						{	
							uint16_t shade = 0;
							
							RollzHLinePlanar( ((const uint16_t *)srcp) + hh * pitch + (planewd - 1 - ww),  
												((uint16_t *)dstp) + hh * pitch + (planewd - 1 - ww), rolldia/2, shade);
						}
						else if(fi->sampleType == stFloat)
						{	
							float shade = 0;
							
							RollzHLinePlanar( ((const float *)srcp) + hh * pitch + (planewd - 1 - ww),  
												((float *)dstp) + hh * pitch + (planewd - 1 - ww), rolldia/2, shade);
						}
				
					}
				} // h= 0
			}

			else if( ( d->rollin && d->dir == 4)  || (!d->rollin && d->dir == 8) ) //, "se")
			{	
				for ( int h = 0; h < dy; h ++)
				{
					int ww = (dy - h) * planewd / planeht;
					ww = ww > planewd -1? planewd -1 : ww;

					int hh = h;
					if ( hh > planeht - 2 || hh < 2) continue;

					vs_bitblt( dstp + hh * stride,stride, srcp + hh * stride,stride, ww * kb, 1);

					int rolldia = 0;
					if( ww < maxroll)
						rolldia = ww;
					else if( ww > planewd - maxroll)
						rolldia = planewd - ww ;
					else
						rolldia = maxroll;

					if(rolldia > ( 8 >> subW))
					{
						if(fi->sampleType == stInteger && nb == 8)
						{	
							uint8_t shade = 0;
						
							RollzHLinePlanar( srcp + hh * pitch + ww ,  
												dstp + hh * pitch + ww, rolldia/2, shade);
						}
						else if(fi->sampleType == stInteger && nb > 8)
						{	
							uint16_t shade = 0;
						
							RollzHLinePlanar( ((const uint16_t *)srcp) + hh * pitch + ww ,  
												((uint16_t *)dstp) + hh * pitch + ww, rolldia/2, shade);
						}
						else if(fi->sampleType == stFloat)
						{	
							float shade = 0.0f;
						
							RollzHLinePlanar( ((const float *)srcp) + hh * pitch + ww ,  
												((float *)dstp) + hh * pitch + ww, rolldia/2, shade);
						}
				
					}
					
				}
			}
			else if( ( d->rollin && d->dir == 2)  || (!d->rollin && d->dir ==6) ) // "ne"
			{	
				for ( int h = 0; h < dy; h ++)
				{
					int ww = ((dy - h) * planewd) / planeht;
					ww = ww > planewd -1? planewd -1 : ww < 0 ? 0 : ww;

					int hh = planeht - h;
					if ( hh > planeht - 2 || hh < 2) continue;

					vs_bitblt(dstp + hh * stride, stride, srcp + hh * stride,stride, ww * kb, 1);

					int rolldia = 0;
					if( ww < maxroll)
						rolldia = ww;
					else if( ww > planewd - maxroll)
						rolldia = planewd - ww ;
					else
						rolldia = maxroll;

					if(rolldia > ( 8 >> subW))
					{
						if(fi->sampleType == stInteger && nb == 8)
						{	
							uint8_t shade = 0;
						
							RollzHLinePlanar(srcp + hh * pitch + ww ,  
												dstp + hh * pitch + ww, rolldia/2, shade);
						}
						else if(fi->sampleType == stInteger && nb > 8)
						{	
							uint16_t shade = 0;
						
							RollzHLinePlanar( ((const uint16_t *)srcp) + hh * pitch + ww ,  
												((uint16_t *)dstp) + hh * pitch + ww, rolldia/2, shade);
						}

						else if(fi->sampleType == stFloat)
						{	
							float shade = 0.0f;
						
							RollzHLinePlanar( ((const float *)srcp) + hh * pitch + ww ,  
												((float *)dstp) + hh * pitch + ww, rolldia/2, shade);
						}
				
					}
					
				}
			}

			else if( ( d->rollin && d->dir == 8)  || (!d->rollin && d->dir == 4) ) // "nw")
			{
				for ( int h = 0; h < dy; h ++)
				{
					int ww = (h) * planewd / planeht;
					ww = ww > planewd - 1? planewd - 1 : ww < 0 ? 0 : ww;

					int hh = planeht - dy + h; //planeht - h;
					if(hh > planeht - 2 || hh < 2) continue;;

					vs_bitblt(dstp + hh * stride + (planewd - 1 - ww ) * kb,stride,  
								srcp + hh * stride + (planewd - 1 - ww) * kb,stride, ww * kb, 1);

					int rolldia = 0;
					if( ww < maxroll)
						rolldia = ww;
					else if( ww > planewd - maxroll)
						rolldia = planewd - ww ;
					else
						rolldia = maxroll;

					if(rolldia > ( 8 >> subW))
					{
						if(fi->sampleType == stInteger && nb == 8)
						{	
							uint8_t shade = 0;
						
							RollzHLinePlanar( srcp + hh * pitch + planewd - 1 -ww ,  
												dstp + hh * pitch + planewd - 1 -ww, rolldia/2, shade);
						}
						else if(fi->sampleType == stInteger && nb > 8)
						{	
							uint16_t shade = 0;
						
							RollzHLinePlanar( ((const uint16_t *)srcp) + hh * pitch + planewd - 1 -ww ,  
												((uint16_t *)dstp) + hh * pitch + planewd - 1 -ww, rolldia/2, shade);
						}

						else if(fi->sampleType == stFloat)
						{	
							float shade = 0;
						
							RollzHLinePlanar( ((const float *)srcp) + hh * pitch + planewd - 1 -ww ,  
												((float *)dstp) + hh * pitch + planewd - 1 -ww, rolldia/2, shade);
						}
				
					}
					
				}
			}	// d->dir
			}


		}
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC rollFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    RollData *d = (RollData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC rollCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    RollData d;
    RollData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Roll: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Roll: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Roll: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir = int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err)
		d.dir = 1;

	if( d.dir < 1 || d.dir > 8)
	{
		vsapi->setError(out, "Roll: valid values for dir are 1 to 8.  1 for north to 8 for nw in clockwise direction");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	temp = !! vsapi->propGetInt(in, "rollin", 0, &err);
	if( err)
		d.rollin = true;
	else if(temp == 0)
		d.rollin = false;
	else
		d.rollin = true;
	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (RollData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Roll", rollInit, rollGetFrame, rollFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "roll", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Roll", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;rollin:int;opt;", rollCreate, 0, plugin);
}
*/