/******************************************************************************
vcTrans.Scratch filter plugin function for vapoursynth by V.C.Mohan
The A Clip  unmasks a stationary B Clip clip in the specified manner of scratcing a scratch card .
 The clips should have identical lengths, height, width and same pixel type 

5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int style;		// scratch style 1;TV noise, 2:rain, 3:dust,
					// 4:laser beams, 5:fountain 6: vertical lines
					// 7: horizontal lines
	int vtransStart;
	int vtransEnd;
	int numFrames;	
	int nrandom, *nrandomx, * nrandomy;
	int span;
	bool nrand;
}ScratchData;
//*************************************************
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC scratchInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    ScratchData *d = (ScratchData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	int width = d->vi.width;

	int height = d->vi.height; 

	d->nrand = false;

	int nrand, wmod, hmod, nyrand, nxrand;

	if(d->style  == 1 )	// TV
	{
		d->span = 4;
		
	}

	else if(d->style == 2 )	// dust
	{ 
		d->span = 1;	
	}

	else if(d->style == 3  )	// rain
	{
		d->span = 6;	
	}	

	if(d->style < 4)		
	{
		nyrand = height - d->span;
		nxrand = width - d->span;
		nrand = nyrand * nxrand;		
		wmod = nxrand ;
		hmod = nyrand ;
	}
	else if(d->style == 4 || d->style == 5 )	// laser or fountain
	{
		d->span = 0;
		nrand = width/2;
		nyrand = height;
		nxrand = width/2 - 1;
		wmod = (width / 2 );
		hmod = 2 * (height );
	}
	

	else if(d->style == 6  )	// vertical lines
	{
		d->span = 0;
		nrand = width;
		nyrand = height;
		nxrand = width;
		wmod = (width);
		hmod = height;
	}
	else if(d->style == 7  )	// horizontal lines
	{
		d->span = 0;
		nrand = height;
		nyrand = height;
		nxrand = width;
		wmod = width;
		hmod = height;
	}
	d->nrandom = nrand;
	d->nrandomy = (int*)vs_aligned_malloc<int>(sizeof(int) * nyrand, 32);// new int[nyrand];
	d->nrandomx = (int*)vs_aligned_malloc<int>(sizeof(int) * nxrand, 32);//new int [nxrand];
		// ensure no repeat numbers in this random set
		for( int i = 0, j = 0; i < nyrand; i ++)
		{
			while (true)
			{
				d->nrandomy[i] = (rand() % hmod);	// & 0xfffffffe;			
				for(j=0;j<i; j++)
				{
					if(d->nrandomy[i]== d->nrandomy[j])	// we got this number before
					break;
				}			
				if(j==i)	// we reached end . so unique number. accept
					break;
			}
		}

		for(int  i = 0, j = 0; i < nxrand; i ++)
		{
			while (true)
			{
				d->nrandomx[i] = (rand() % wmod);
				
				for( j=0;j<i; j++)
				{
					if(d->nrandomx[i]== d->nrandomx[j])	// we got this number before
					break;
				}
			
				if(j==i)	// we reached end . so unique number. accept
					break;
			}
		}

	d->nrand = true;
}
/***************************************************************/	
static const VSFrameRef *VS_CC scratchGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     ScratchData *d = (ScratchData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		VSFrameRef * dst =  vsapi->copyFrame ( srca, core) ;

		const uint8_t * srpa[] = {NULL, NULL, NULL};
		const uint8_t * srpb[] = {NULL, NULL, NULL};
		uint8_t * dsp[] = {NULL, NULL, NULL};
		int pitch[] = {0, 0, 0};;

		int subH = (fi->colorFamily == cmYUV) ? fi->subSamplingH : 0;
		int subW = (fi->colorFamily == cmYUV) ? fi->subSamplingW : 0;
		int andH = (1 << subH ) - 1;
		int andW = (1 << subW ) - 1;
		// as all plane pointers are needed for the scratch we get them here
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *  dstp = vsapi->getWritePtr(dst, plane);
		//	int ht = vsapi->getFrameHeight(srca, plane);
		//	int wd = vsapi->getFrameWidth(srca, plane);
		//	int pitch = stride / kb;	
			srpa[plane] = srcpa;
			srpb[plane] = srcpb;
			dsp[plane] = dstp;
			pitch[plane] = stride/ kb;
		}		
		if( fi->sampleType == stInteger && nb == 8)
		{
			const uint8_t * sp = srpb[0];
			const uint8_t * spU = srpb[1];
			const uint8_t * spV = srpb[2];

			uint8_t * dp = dsp[0];
			uint8_t * dpU = dsp[1];
			uint8_t * dpV = dsp[2];

			if(d->style  == 1)		// "tv"
			{

				int npoints = ( (bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy  random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);
					int npx = (shiftx + np) % ( width - d->span);

					int xspan = 1 + (d->nrandomx[npx] % d->span); 
					int yspan = 1 + (d->nrandomy[npy] % d->span);	// spot area dimensions

					for( int h = 0; h < yspan; h ++)
					{					
						for ( int w = 0; w < xspan; w ++)
						{							
							dp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)] 
									= sp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)];

							if( fi->numPlanes == 1 )
								continue;		

							if( (h & andH) == 0 && (w & andW ) == 0 )
							{
								dpU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW) ] 
									= spU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)];

								dpV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ] 
									= spV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ];
							}							
						}
					}	
				}
			}

			else if(d->style  == 2)	// dust
			{

						// copy random color values from right Frame to corresponding point
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy a random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{

						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{

						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					dp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx] ) ] 
							= sp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx]) ];

					if ( fi->numPlanes > 1)
					{
					
						dpU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];

						dpV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];
					}								
				}
			}
		

			else if(d->style  == 3)		// rain
			{
				int npoints = ( ( bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;				

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}

					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					for( int xy = 0; xy < d->span; xy ++)
					{						
						
						dp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy) ] 
								= sp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy)];

						if ( fi->numPlanes >> 1)
						{

							if( ( xy  & subH) == 0 && (xy  & subW) == 0  )
							{
								dpU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW) ] 
									= spU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];

								dpV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW)] 
									= spV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];
							}
						
						}				
					}
				}
			}

			else if(d->style  == 4)		// "laser beams "
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i = 0; i < npoints; i ++)		// number of beams
				{
				//	int length = d->nrandomy[i];;// doubled h to cover frame fully

					int wmax = d->nrandomx[i];	// dist of beam end from center of frame. halfed w to remain in frame

					for(int h = 0; h < d->nrandomy[i]; h ++)
					{
						int hh = (h < height ? h : 2 * height - 2 - h);	// end y coordinate of beam

						int w = (h * d->nrandomx[i]) / ( 1 + d->nrandomy[i]);	// avoid div by zero

						
						dp [ (width /2 + w) + hh * pitch[0]]

								= sp [ (width /2 + w) + hh * pitch[0]];

										// mirror image of beam

						dp [ (width /2 - w) + hh * pitch[0]]

								= sp [ (width /2 - w) + hh * pitch[0]];

						if ( fi->numPlanes > 1)
						{

							if((hh & andH) == 0 && (w & andW) == 0  )
							{
								// u v planes// mirror image of beam
								dpU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];

								dpU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];

								dpV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];									

								dpV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];
							}
						}						
					}
				}
			}

			else if(d->style == 5)	//"fountain"
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i=0; i < npoints; i ++)
				{

					int py = 2 + d->nrandomy[i];	// avoid div by zero;

					int px = d->nrandomx[i];;

					float pp = (float(px * px)) / (4 * py);

					for(int w = 0; w < width / 2; w ++)
					{
						
						int h = (int)(height - (py - (w - px) * (w - px) / (4 * pp)));
		
						if(h <= 0 || h >= height)
							continue;
						
						dp[h * pitch[0] + (width/2 - w)  ]	
							= sp[h * pitch[0] + (width/2 - w)  ];
							
						dp[h * pitch[0] + (width/2 + w)  ]							
							= sp[h * pitch[0] + (width/2 + w)  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )
								continue;
									// u and v planes
						dpU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];

						dpU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
						
					}
				}
			}		

			else if(d->style == 6)	// vertical lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);
				// scratch at random vertical lines  
				for (int i = 0; i < npoints; i++)
				{
					int w = d->nrandomx[i];

					for(int h = 0; h < height; h ++)
					{						
						dp [h * pitch[0] + w   ]

							= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
					}
				}
			}

			else if(d->style == 7  )	// horizontal lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for (int i = 0; i < npoints; i++)
				{
					int h = d->nrandomy[i];

					for(int w = 0; w < width; w ++)
					{						
						dp [h * pitch[0] + w  ]

								= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
											
					}
				}

			}

		}

		else if( fi->sampleType == stInteger && nb > 8)
		{
			const uint16_t * sp = (const uint16_t * )srpb[0];
			const uint16_t * spU = (const uint16_t * )srpb[1];
			const uint16_t * spV = (const uint16_t * )srpb[2];

			uint16_t * dp = (uint16_t * )dsp[0];
			uint16_t *  dpU = (uint16_t * )dsp[1];
			uint16_t *  dpV = (uint16_t * )dsp[2];

			if(d->style  == 1)		// "tv"
			{

				int npoints = ( (bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy  random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);
					int npx = (shiftx + np) % ( width - d->span);

					int xspan = 1 + (d->nrandomx[npx] % d->span); 
					int yspan = 1 + (d->nrandomy[npy] % d->span);	// spot area dimensions

					for( int h = 0; h < yspan; h ++)
					{					
						for ( int w = 0; w < xspan; w ++)
						{							
							dp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)] 
									= sp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)];

							if( fi->numPlanes == 1 )
								continue;		

							if( (h & andH) == 0 && (w & andW ) == 0 )
							{
								dpU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW) ] 
									= spU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)];

								dpV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ] 
									= spV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ];
							}							
						}
					}	
				}
			}

			else if(d->style  == 2)	// dust
			{

						// copy random color values from right Frame to corresponding point
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy a random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{

						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{

						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					dp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx] ) ] 
							= sp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx]) ];

					if ( fi->numPlanes > 1)
					{
					
						dpU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];

						dpV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];
					}								
				}
			}
		

			else if(d->style  == 3)		// rain
			{
				int npoints = ( ( bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;				

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}

					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					for( int xy = 0; xy < d->span; xy ++)
					{						
						
						dp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy) ] 
								= sp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy)];

						if ( fi->numPlanes >> 1)
						{

							if( ( xy  & subH) == 0 && (xy  & subW) == 0  )
							{
								dpU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW) ] 
									= spU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];

								dpV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW)] 
									= spV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];
							}
						
						}				
					}
				}
			}

			else if(d->style  == 4)		// "laser beams "
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i = 0; i < npoints; i ++)		// number of beams
				{
				//	int length = d->nrandomy[i];;// doubled h to cover frame fully

					int wmax = d->nrandomx[i];	// dist of beam end from center of frame. halfed w to remain in frame

					for(int h = 0; h < d->nrandomy[i]; h ++)
					{
						int hh = (h < height ? h : 2 * height - 2 - h);	// end y coordinate of beam

						int w = (h * d->nrandomx[i]) / ( 1 + d->nrandomy[i]);	// avoid div by zero

						
						dp [ (width /2 + w) + hh * pitch[0]]

								= sp [ (width /2 + w) + hh * pitch[0]];

										// mirror image of beam

						dp [ (width /2 - w) + hh * pitch[0]]

								= sp [ (width /2 - w) + hh * pitch[0]];

						if ( fi->numPlanes > 1)
						{

							if((hh & andH) == 0 && (w & andW) == 0  )
							{
								// u v planes// mirror image of beam
								dpU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];

								dpU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];

								dpV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];									

								dpV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];
							}
						}						
					}
				}
			}

			else if(d->style == 5)	//"fountain"
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i=0; i < npoints; i ++)
				{

					int py = 2 + d->nrandomy[i];	// avoid div by zero;

					int px = d->nrandomx[i];;

					float pp = (float(px * px)) / (4 * py);

					for(int w = 0; w < width / 2; w ++)
					{
						
						int h = (int)(height - (py - (w - px) * (w - px) / (4 * pp)));
		
						if(h <= 0 || h >= height)
							continue;
						
						dp[h * pitch[0] + (width/2 - w)  ]	
							= sp[h * pitch[0] + (width/2 - w)  ];
							
						dp[h * pitch[0] + (width/2 + w)  ]							
							= sp[h * pitch[0] + (width/2 + w)  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )
								continue;
									// u and v planes
						dpU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];

						dpU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
						
					}
				}
			}		

			else if(d->style == 6)	// vertical lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);
				// scratch at random vertical lines  
				for (int i = 0; i < npoints; i++)
				{
					int w = d->nrandomx[i];

					for(int h = 0; h < height; h ++)
					{						
						dp [h * pitch[0] + w   ]

							= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
					}
				}
			}

			else if(d->style == 7  )	// horizontal lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for (int i = 0; i < npoints; i++)
				{
					int h = d->nrandomy[i];

					for(int w = 0; w < width; w ++)
					{						
						dp [h * pitch[0] + w  ]

								= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
											
					}
				}

			}

		}

		else if( fi->sampleType == stFloat)
		{
			const float * sp = (const float * )srpb[0];
			const float * spU = (const float * )srpb[1];
			const float * spV = (const float * )srpb[2];

			float * dp = (float * )dsp[0];
			float *  dpU = (float * )dsp[1];
			float *  dpV = (float * )dsp[2];

			if(d->style  == 1)		// "tv"
			{

				int npoints = ( (bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy  random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);
					int npx = (shiftx + np) % ( width - d->span);

					int xspan = 1 + (d->nrandomx[npx] % d->span); 
					int yspan = 1 + (d->nrandomy[npy] % d->span);	// spot area dimensions

					for( int h = 0; h < yspan; h ++)
					{					
						for ( int w = 0; w < xspan; w ++)
						{							
							dp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)] 
									= sp[(d->nrandomy[npy] + h) * pitch[0] + (d->nrandomx[npx] + w)];

							if( fi->numPlanes == 1 )
								continue;		

							if( (h & andH) == 0 && (w & andW ) == 0 )
							{
								dpU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW) ] 
									= spU[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)];

								dpV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ] 
									= spV[((d->nrandomy[npy] + h) >> subH) * pitch[1] + ((d->nrandomx[npx] + w) >> subW)  ];
							}							
						}
					}	
				}
			}

			else if(d->style  == 2)	// dust
			{

						// copy random color values from right Frame to corresponding point
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;
					// at random points in the frame copy a random sized spots from  right frame

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{

						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{

						if( np% (width - d->span) == 0)

							shifty ++;
					}


					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					dp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx] ) ] 
							= sp[(d->nrandomy[npy]) * pitch[0] + (d->nrandomx[npx]) ];

					if ( fi->numPlanes > 1)
					{
					
						dpU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spU[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];

						dpV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx] ) >> subW) ] 
									= spV[((d->nrandomy[npy]) >> subH) * pitch[1] + ((d->nrandomx[npx]) >> subW) ];
					}								
				}
			}
		

			else if(d->style  == 3)		// rain
			{
				int npoints = ( ( bn + 1) * d->nrandom) / (d->overlap + 1);

				int shiftx = 0, shifty = 0;				

				for( int np  = 0; np < npoints; np ++)
				{
					if (height > width)
					{
						if( np % (height - d->span) == 0)

							shiftx ++;
					} 
					else
					{
						if( np% (width - d->span) == 0)

							shifty ++;
					}

					int npy = (shifty + np) % (height - d->span);

					int npx = (shiftx + np) % ( width - d->span);
					
					for( int xy = 0; xy < d->span; xy ++)
					{						
						
						dp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy) ] 
								= sp[(d->nrandomy[npy] + xy) * pitch[0] + (d->nrandomx[npx] + xy)];

						if ( fi->numPlanes >> 1)
						{

							if( ( xy  & subH) == 0 && (xy  & subW) == 0  )
							{
								dpU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW) ] 
									= spU[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];

								dpV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy ) >> subW)] 
									= spV[((d->nrandomy[npy] + xy) >> subH) * pitch[1] + ((d->nrandomx[npx] + xy) >> subW) ];
							}
						
						}				
					}
				}
			}

			else if(d->style  == 4)		// "laser beams "
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i = 0; i < npoints; i ++)		// number of beams
				{
				//	int length = d->nrandomy[i];;// doubled h to cover frame fully

					int wmax = d->nrandomx[i];	// dist of beam end from center of frame. halfed w to remain in frame

					for(int h = 0; h < d->nrandomy[i]; h ++)
					{
						int hh = (h < height ? h : 2 * height - 2 - h);	// end y coordinate of beam

						int w = (h * d->nrandomx[i]) / ( 1 + d->nrandomy[i]);	// avoid div by zero

						
						dp [ (width /2 + w) + hh * pitch[0]]

								= sp [ (width /2 + w) + hh * pitch[0]];

										// mirror image of beam

						dp [ (width /2 - w) + hh * pitch[0]]

								= sp [ (width /2 - w) + hh * pitch[0]];

						if ( fi->numPlanes > 1)
						{

							if((hh & andH) == 0 && (w & andW) == 0  )
							{
								// u v planes// mirror image of beam
								dpU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];

								dpU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spU [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];

								dpV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 + w) >> subW) + (hh >> subH) * pitch[1]];									

								dpV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]]
								= spV [ ((width /2 - w) >> subW) + (hh >> subH) * pitch[1]];
							}
						}						
					}
				}
			}

			else if(d->style == 5)	//"fountain"
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for(int i=0; i < npoints; i ++)
				{

					int py = 2 + d->nrandomy[i];	// avoid div by zero;

					int px = d->nrandomx[i];;

					float pp = (float(px * px)) / (4 * py);

					for(int w = 0; w < width / 2; w ++)
					{
						
						int h = (int)(height - (py - (w - px) * (w - px) / (4 * pp)));
		
						if(h <= 0 || h >= height)
							continue;
						
						dp[h * pitch[0] + (width/2 - w)  ]	
							= sp[h * pitch[0] + (width/2 - w)  ];
							
						dp[h * pitch[0] + (width/2 + w)  ]							
							= sp[h * pitch[0] + (width/2 + w)  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )
								continue;
									// u and v planes
						dpU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];

						dpU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]
							= spU[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 - w) >> subW)  ];
							
						dpV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ]	
							= spV[(h >> subH) * pitch[1] + ((width/2 + w) >> subW)  ];
						
					}
				}
			}		

			else if(d->style == 6)	// vertical lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);
				// scratch at random vertical lines  
				for (int i = 0; i < npoints; i++)
				{
					int w = d->nrandomx[i];

					for(int h = 0; h < height; h ++)
					{						
						dp [h * pitch[0] + w   ]

							= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
					}
				}
			}

			else if(d->style == 7  )	// horizontal lines
			{
				int npoints = ((bn + 1) * d->nrandom) / (d->overlap + 1);

				for (int i = 0; i < npoints; i++)
				{
					int h = d->nrandomy[i];

					for(int w = 0; w < width; w ++)
					{						
						dp [h * pitch[0] + w  ]

								= sp [h * pitch[0] + w  ];

						if((h & andH) != 0 || (w & andW) != 0 || fi->numPlanes == 1 )

							continue;

									// u and v planes

						dpU[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spU[(h >> subH) * pitch[1] + ((w >> subW)  )];

						dpV[(h >> subH) * pitch[1] + ((w >> subW)  )]

						= spV[(h >> subH) * pitch[1] + ((w >> subW)  )];
											
					}
				}

			}

		}
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC scratchFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    ScratchData *d = (ScratchData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	vs_aligned_free(d->nrandomx);
	vs_aligned_free(d->nrandomy);
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC scratchCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    ScratchData d;
    ScratchData *data;
    int err;
//	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Scratch: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Scratch: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Scratch: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.style = int64ToIntS(vsapi->propGetInt(in, "style", 0, &err));
	if( err)
		d.style = 3;

	if( d.style < 1 || d.style > 7)
	{
		vsapi->setError(out, "Scratch: Style valid numbers 1 to 7.  1: TV, 2: dust, 3: rain 4: laser 5: fountain 6: v lines 7: H lines ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (ScratchData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Scratch", scratchInit, scratchGetFrame, scratchFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "scratch", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Scratch", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;", scratchCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------


