/******************************************************************************
Trans.Shuffle filter plugin function for vapoursynth by V.C.Mohan
The A and B Clips  move similar to the movement 
of a deck of cards while shuffling.
 The clips should have identical lengths, height, width and same pixel type 
5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int dir;		// shuffle direction
	int vtransStart;
	int vtransEnd;
	int numFrames;	
	
}ShuffleData;
/*************************************************
 
 *************************************************/
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC shuffleInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    ShuffleData *d = (ShuffleData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

}

	
/***************************************************************/	
static const VSFrameRef *VS_CC shuffleGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     ShuffleData *d = (ShuffleData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		VSFrameRef * dst =  vsapi->newVideoFrame (fi, width, height, srca, core) ;
	
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *  dstp = vsapi->getWritePtr(dst, plane);
			int ht = vsapi->getFrameHeight(srca, plane);
			int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / kb;
			int deltay =((bn + 1) * ht)/(d->overlap + 1);
			int deltax =((bn + 1) * wd)/(d->overlap + 1);
			if ( d->dir == 2 || d->dir == 4)
			{
				if(deltax <= 1)
				{
					vsapi->freeFrame(srcb);
					return srca;
				}

				else if(deltax >=  wd - 1)
				{
					vsapi->freeFrame(srca);
					return srcb;
				}
			}
			else // dir = 1 or 3
			{
				if(deltay <= 1)
				{
					vsapi->freeFrame(srcb);
					return srca;
				}

				else if(deltay >=  ht - 1)
				{
					vsapi->freeFrame(srca);
					return srcb;
				}
			}


			bool flipped = false;

			if(bn >= d->overlap / 2)
			{
				flipped = true;
				deltay -= ht / 2;
				deltax -= wd / 2;
			}
		
			int aFramex, aFramey, bFramex, bFramey;
			int dax, dbx, day, dby ;
			int awd, aht, bht, bwd;

			if ( d->dir == 1  ) 
			{
					// A Frame moves north, while B Frame from underneath moves down
					//gradually covering A Frame till half time. Then B frame moves up
					// while A hrame slides in below B frame
				if(!flipped)
				{
					// copy A frame from these coord and width, height
					aFramex =0;
					aFramey = deltay;
					awd = wd;				
					aht= ht - deltay;
					// position of A on dst
					dax = 0;
					day = 0;
					// copy B frame from these coord and width, height
					bFramex = 0;
					bFramey = ht - 2 * deltay;
					bht = deltay;
					bwd = wd;
					// position of B on dst
					dbx = 0;
					dby = ht - deltay;
				}
				else	//if(flipped)
				{
					aFramex = 0;
					aFramey = ht / 2 - deltay;
					awd = wd;				
					aht = ht/2 - deltay;
					dax = 0;
					day = 0;
					bFramex = 0;
					bFramey = 0;
					bht = ht / 2 + deltay;
					bwd = wd;
					dbx = 0;
					dby = ht / 2 - deltay;
				}
			}

			else if ( d->dir == 3  )		
			{
						// A Frame moves down halfway, while B Frame from underneath moves up
						//gradually till half frames are seen. Then B frame moves down
						// while A frame slides up below B frame
				if(!flipped)
				{
					aFramex = 0;
					aFramey = 0;
					awd = wd;				
					aht = ht - deltay;
					dax = 0;
					day = deltay;
					bFramex=0;
					bFramey = deltay;
					bht = deltay;
					bwd = wd;
					dbx = 0;
					dby = 0;
				}
				else	//if(flipped)
				{
					aFramex = 0;
					aFramey = 2 * deltay;
					awd = wd;				
					aht = ht / 2 - deltay;
					dax = 0;
					day = ht / 2 + deltay;
					bFramex = 0;
					bFramey = ht / 2 - deltay;
					bht = ht / 2 + deltay;
					bwd = wd;
					dbx = 0;
					dby = 0;
				}				
			}

			else if(d->dir == 4)	// "left")
			{
						// A Frame moves left halfway while B Frame 
						// from underneath moves Right gradually till two halves
						// are seen. Then B frame moves left while A frame slides
						// right underneath
				if(!flipped)
				{
					aFramex = deltax;
					aFramey = 0;
					awd = wd - deltax;				
					aht = ht;
					dax = 0;
					day = 0;
					bFramex = wd - 2 * deltax;
					bFramey = 0;
					bht = ht;
					bwd = deltax;
					dbx = wd - deltax;
					dby = 0;
				}
				else	//if(flipped)
				{
					aFramex = wd / 2 - deltax;
					aFramey = 0;
					awd = wd / 2 - deltax;				
					aht = ht;
					dax = 0;
					day = 0;
					bFramex = 0;
					bFramey = 0;
					bht = ht;
					bwd = wd / 2 + deltax;
					dbx = wd / 2 - deltax;
					dby = 0;
				}				
			}

			else	//right
			{
						// A Frame moves right halfway while B Frame 
						// from underneath moves Left gradually till two halves
						// are seen. Then B frame moves Right while A frame slides
						// leftt underneath
				if(!flipped)
				{
					aFramex = 0;
					aFramey = 0;
					awd = wd - deltax;				
					aht = ht;
					dax = deltax;
					day = 0;
					bFramex = deltax;
					bFramey = 0;
					bht = ht;
					bwd = deltax;
					dbx = 0;
					dby = 0;
				}
				else	//if(flipped)
				{
					aFramex = 2 * deltax;
					aFramey = 0;
					awd=wd/2 - deltax;				
					aht = ht;
					dax = wd / 2 + deltax;
					day = 0;
					bFramex = wd / 2 - deltax;
					bFramey = 0;
					bht = ht;
					bwd = wd / 2 + deltax;
					dbx = 0;
					dby = 0;
				}				
			}
				// copy the parts of A  on to dst
			vs_bitblt(dstp + day * stride + kb * dax , stride,
					srcpa + aFramey * stride + kb * aFramex , stride,
					awd * kb, aht);
			// copy the parts of  B frame on to dst
			vs_bitblt(dstp + dby * stride + kb * dbx , stride,
					srcpb + bFramey * stride + kb * bFramex, stride,
					bwd * kb, bht);


		
		}

		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
	
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC shuffleFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    ShuffleData *d = (ShuffleData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC shuffleCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    ShuffleData d;
    ShuffleData *data;
    int err;
//	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Shuffle: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Shuffle: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Shuffle: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir = int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err)
		d.dir = 1;

	if( d.dir < 1 || d.dir > 4)
	{
		vsapi->setError(out, "Shuffle: dir valid numbers 1 to 4.  1: up, 2: right, 3: down 4: left ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (ShuffleData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Shuffle", shuffleInit, shuffleGetFrame, shuffleFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "shuffle", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Shuffle", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", shuffleCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------


