/******************************************************************************
Trans.Slide filter plugin function for vapoursynth by V.C.Mohan
B Clip gradually Masks  a stationary A Clip or A clip slides out unmasking a stationary
B clip  in the direction specified.
 The clips should have identical lengths, height, width and same pixel type 
5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int dir;		// slide dir 1 n 2;ne ,3 e, 4:se,5 s, 6:sw,7 w, 8 :nw
	
	bool slidein;
	int vtransStart;
	int vtransEnd;
	int numFrames;	
	
}SlideData;
//------------------------------------------------------------------
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC slideInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    SlideData *d = (SlideData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

}

//--------------------------------------------------------------------------------	
static const VSFrameRef *VS_CC slideGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     SlideData *d = (SlideData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		VSFrameRef * dst = d->slidein ?  vsapi->copyFrame ( srca, core) :vsapi->copyFrame (srcb, core) ;
		const VSFrameRef *src = d->slidein ? srcb: srca;

		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t *srcp = vsapi->getReadPtr(src, plane);
			
			uint8_t * dstp = vsapi->getWritePtr(dst, plane);
			int stride = vsapi->getStride(srca, plane);
			int pitch = stride / kb;
			int planeht = vsapi->getFrameHeight(srca, plane);
			int planewd = vsapi->getFrameWidth(srca, plane);
			int subH = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingH : 0;
			int subW = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
			int andH = (1 << subH) -1;
			int andW = (1 << subW) -1;

			int deltax = ((bn+1)* planewd)/(d->overlap+1) ;
			int deltay = ((bn+1)* planeht)/(d->overlap+1) ;
			if ( ! d-> slidein)
			{
				deltax = planewd - deltax;
				deltay = planeht - deltay;
			}
			int dstX, dstY, wht, wwd, srcx, srcy ; // positions and widths of  frames at this transition

			if( (d->slidein && d->dir == 1) || ( !d->slidein && d->dir == 5))	//		"north"	
			{
					// B Frame moves north gradually covering LeftFrame
				dstX =0;
				dstY = planeht-deltay;
				wht = deltay;
				wwd = planewd;
				srcx=0;
				srcy=0;
			}

			else if( (d->slidein && d->dir == 5) || ( !d->slidein && d->dir == 1))	// south
		
			{
						// src moves south gradually covering LeftFrame
					dstX =0;
					dstY = 0;
					wht = deltay;
					wwd = planewd;
					srcx=0;
					srcy=planeht-deltay;
				
			}

			else if( (d->slidein && d->dir == 3) || ( !d->slidein && d->dir == 7))	//east
			{
					// src moves east gradually covering LeftFrame
				dstX =0;
				dstY = 0;
				wht = planeht;
				wwd = deltax;
				srcx = planewd - deltax;
				srcy=0;
			
			}

			else if( (d->slidein && d->dir == 7) || ( !d->slidein && d->dir == 3))	//west

			{
				// src moves west masking LeftFrame
				dstX =planewd-deltax;
				dstY = 0;
				wht = planeht;
				wwd = deltax;
				srcx=0;
				srcy=0;
			
			}

			else if( (d->slidein && d->dir == 2) || ( !d->slidein && d->dir == 6))	//ne

			{
					// src moves north east gradually masking LeftFrame
				dstX =0;
				dstY = planeht-deltay;
				wht = deltay;
				wwd = deltax;
				srcx= planewd - deltax;
				srcy=0;
			
			}

			else if( (d->slidein && d->dir == 4) || ( !d->slidein && d->dir == 8))	// se
			{
					// src moves south east covering LeftFrame
				dstX =0;
				dstY =0;
				wht = deltay;
				wwd = deltax;
				srcx=planewd-deltax;
				srcy=planeht-deltay;
			
			}

			else if( (d->slidein && d->dir == 8) || ( !d->slidein && d->dir == 4))	// nw

			{
					// src moves north west gradually masking LeftFrame
				dstX =planewd-deltax;
				dstY = planeht-deltay;
				wht = deltay;
				wwd = deltax;
				srcx=0;
				srcy=0;
			
			}

			else if( (d->slidein && d->dir == 6) || ( !d->slidein && d->dir == 2))	// sw

			{
					// src moves south west gradually masking LeftFrame
				dstX = planewd-deltax;
				dstY = 0;
				wht = deltay;
				wwd = deltax;
				srcx=0;
				srcy=planeht-deltay;
			
			}

			else if( d->dir == 0)	// from center of frame
			{
					// B emerges from center of A Frame in all directions
				dstX =(planewd-deltax)/2;
				dstY =(planeht- deltay)/2;
				wht = deltay;
				wwd = deltax;
				srcx = (planewd-deltax)/2;
				srcy = (planeht-deltay)/2;				
			}

			if (wwd==0 || wht==0)
			{
				vsapi->freeFrame (srcb);			
				return srca; // no work to be done
			}
	
			vs_bitblt( dstp + dstY * stride + dstX * kb, stride,
					srcp + srcy * stride + srcx * kb, stride, 
					wwd * kb, wht);
				
			
				
		}	// plane

		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
	
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC slideFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    SlideData *d = (SlideData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC slideCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    SlideData d;
    SlideData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Slide: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Slide: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Slide: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir = int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err)
		d.dir = 1;

	if(  d.dir < 0 || d.dir > 8 )
	{
		vsapi->setError(out, "Slide: dir valid numbers 0 for central, and 1 to 8.  1:north to 8: nw in clockwise");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}	

	temp = !! vsapi->propGetInt(in, "slidein", 0, &err);
	if( err)
		d.slidein = true;

	else if(  temp < 0 || temp > 1 )
	{
		vsapi->setError(out, "Slide: value for slidein can be either 1 or zero ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}

	else
	{

		if(temp == 0)
			d.slidein = false;
		else
			d.slidein = true;
	}
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (SlideData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Slide", slideInit, slideGetFrame, slideFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "slide", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Slide", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;slidein:int:opt;", slideCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
