/******************************************************************************
Trans.Sprite filter function of plugin for vapoursynth by V.C.Mohan
The B Clip clip rotating into view pushing out A Clip in the direction specified.
This appears as rotating 4 sided billboard
 The clips should have identical lengths, height, width and same pixel type 

5 Aug 2015  3 june 2021, 17 July 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 

*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int dir;		// sprite dir 1;up ,2 left, 3 down, 4 right
	
	int vtransStart;
	int vtransEnd;
	int numFrames;	
	
}SpriteData;
//------------------------------------------------------------------
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC spriteInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    SpriteData *d = (SpriteData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

}

//--------------------------------------------------------------------------------	
static const VSFrameRef *VS_CC spriteGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     SpriteData *d = (SpriteData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		// B frame will be revealed during transition
		
		int nbytes = fi->bytesPerSample;
		int nbits = fi->bitsPerSample;	

		VSFrameRef * dst = vsapi->newVideoFrame(fi, width, height, srca, core);
		uint8_t blackBGR[] = { 0,0,0,0 };
		uint8_t blackYUV[] = { (uint8_t)16,(uint8_t)128,(uint8_t)128,0 };
		uint8_t* blackCol = fi->colorFamily == cmRGB ? blackBGR : blackYUV;
		

		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t *srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t *srcpb = vsapi->getReadPtr(srcb, plane);
			
			uint8_t * dstp = vsapi->getWritePtr(dst, plane);
			int stride = vsapi->getStride(srca, plane);
			int pitch = stride / nbytes;
			int planeht = vsapi->getFrameHeight(srca, plane);
			int planewd = vsapi->getFrameWidth(srca, plane);
			int subH = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingH : 0;
			int subW = plane > 0 && fi->colorFamily == cmYUV ? fi->subSamplingW : 0;
			int andH = (1 << subH) -1;
			int andW = (1 << subW) -1;

			int deltax = (((bn+1)* planewd)/(d->overlap+1)) & 0xfffffffe;
			int deltay = (((bn+1)* planeht)/(d->overlap+1)) & 0xfffffffe;

			if ( ((d->dir == 1 || d->dir == 3) && deltay < (4 >> subH) )
				|| ((d->dir == 2 || d->dir == 4) && deltax < (4 >> subW) ) )
			{			
				vsapi->freeFrame(srcb);
				return srca;
			}			
			else if ( ((d->dir == 1 || d->dir == 3) && deltay > planeht - (4 >> subH) )
				|| ((d->dir == 2 || d->dir == 4) && deltax > planewd - (4 >> subW) ) )
			{
			
				vsapi->freeFrame(srca);
				return srcb;
			}

			if (d->dir == 1 || d->dir == 2)
			{
				deltay = planeht - deltay;
				deltax = planewd - deltax;
				srcpa = vsapi->getReadPtr(srcb, plane);
				srcpb = vsapi->getReadPtr(srca, plane);
			}
		
			if(d->dir == 4 || d->dir == 2 ) // right or left
			{
				// only width resizing
				float factorb = planewd / (float)deltax;
				float factora = planewd / (float)(planewd - deltax);

				if( fi->sampleType == stInteger && nbits == 8)
				{
					uint8_t * dp = dstp;
					const uint8_t *spa = srcpa;
					const uint8_t *spb = srcpb;	

					HFastResizerPlanar(spa, dp,
						planeht, planewd, planeht, planewd - deltax, pitch,
						pitch, false);
					HFastResizerPlanar(spb, dp + planewd - deltax,
						planeht, planewd, planeht, deltax, pitch,
						pitch, false);					
				}	// integer nbits = 8

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *)dstp;
					const uint16_t *spa = (const uint16_t *)srcpa;
					const uint16_t *spb = (const uint16_t *)srcpb;

					HFastResizerPlanar(spa, dp,
						planeht, planewd, planeht, planewd - deltax, pitch,
						pitch, false);
					HFastResizerPlanar(spb, dp + planewd - deltax,
						planeht, planewd, planeht, deltax, pitch,
						pitch, false);
					
				}	

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *)dstp;
					const float *spa = (const float *)srcpa;
					const float *spb = (const float *)srcpb;

					HFastResizerPlanar(spa, dp,
						planeht, planewd, planeht, planewd - deltax, pitch,
						pitch, false);
					HFastResizerPlanar(spb, dp + planewd - deltax,
						planeht, planewd, planeht, deltax, pitch,
						pitch, false);
				}	// float			

			
				// blacken sides in a tapered form to give depth
				int tapermax = planewd/10;
				int meetw = planewd - deltax; // d->dir == 2 ? deltax : planewd - deltax;// wcoordinate where the two frames meet
			
				// the smaller the frame size (reduced) more is taper	
				int taper1 = ((planewd - meetw) * tapermax) / planewd;
				int taper2 = tapermax - taper1;

				if( fi->sampleType == stInteger && nbits == 8)
				{
					uint8_t * dp = dstp;
					
					for(int w = meetw; w >= 0; w--)
					{
						for(int h=0; h< ((meetw - w) * taper1) / meetw; h ++)
						{
						
							*(dp + h * pitch + w) = blackCol[plane];
							*(dp + (planeht - 1 - h) * pitch + w ) = blackCol[plane];
						}
					}

					for(int w = meetw; w < planewd; w ++)
					{
						for(int h = 0; h < ((w - meetw) * taper2) / (planewd - meetw); h++)
						{
						
							*(dp + h * pitch + w) = blackCol[plane];
							*(dp + (planeht - 1 - h) * pitch + w ) = blackCol[plane];
						}
					}
				} // integer nbits = 8

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *)dstp;
					
					for(int w = meetw; w >= 0; w--)
					{
						for(int h = 0; h< ((meetw - w) * taper1) / meetw; h++)
						{
						
							*(dp + h * pitch + w) = (uint16_t)(blackCol[plane] << (nbits - 8));
							*(dp + (planeht - 1 - h) * pitch + w ) = (uint16_t)(blackCol[plane] << (nbits - 8));
						}
					}

					for(int w = meetw; w < planewd; w ++)
					{
						for(int h = 0; h < ((w - meetw) * taper2) / (planewd - meetw); h++)
						{
						
							*(dp + h * pitch + w) = (uint16_t)(blackCol[plane] << (nbits - 8));
							*(dp + (planeht - 1 - h) * pitch + w ) = (uint16_t)(blackCol[plane] << (nbits - 8));
						}
					}
				} // integer 

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *)dstp;
					
					for(int w = meetw; w >= 0; w --)
					{
						for(int h = 0; h < ((meetw - w) * taper1) / meetw; h++)
						{
						
							*(dp + h * pitch + w) = 0;
							*(dp + (planeht - 1 - h) * pitch + w ) = 0;
						}
					}

					for(int w = meetw; w < planewd; w++)
					{
						for(int h = 0; h < ((w - meetw) * taper2) / (planewd - meetw);h ++)
						{
						
							*(dp + h * pitch + w) = 0;
							*(dp + (planeht - 1 - h) * pitch + w ) = 0;
						}
					}
				} // float
				// rgb or plane = 0
			}	// dir = 2 or 4

			else if( d->dir == 1 || d->dir == 3)	// up or down only vertical movement
			{				
				if (fi->sampleType == stInteger && nbytes == 1)
				{
					uint8_t* dp = dstp;
					const uint8_t* spa = srcpa;
					const uint8_t* spb = srcpb;

					VFastResizerPlanar(spa, dp + deltay * pitch,
						planeht, planewd, planeht - deltay, planewd, pitch,
						pitch, false);
					VFastResizerPlanar(spb, dp,
						planeht, planewd, deltay, planewd, pitch,
						pitch, false);
				}	// integer nbits = 8

				else if (fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t* dp = (uint16_t*)dstp;
					const uint16_t* spa = (const uint16_t*)srcpa;
					const uint16_t* spb = (const uint16_t*)srcpb;

					VFastResizerPlanar(spa, dp + deltay * pitch,
						planeht, planewd, planeht - deltay, planewd, pitch,
						pitch, false);
					VFastResizerPlanar(spb, dp,
						planeht, planewd, deltay, planewd, pitch,
						pitch, false);
				}	// integer nbytes 2

				else if (fi->sampleType == stFloat)
				{
					float* dp = (float*)dstp;
					const float* spa = (const float*)srcpa;
					const float* spb = (const float*)srcpb;

					VFastResizerPlanar(spa, dp + deltay * pitch,
						planeht, planewd, planeht - deltay, planewd, pitch,
						pitch, false);
					VFastResizerPlanar(spb, dp,
						planeht, planewd, deltay, planewd, pitch,
						pitch, false);
				}	// float				
				
				
				// blacken sides in a tapered form to give depth
				int tapermax = planeht/10;
				int meeth = deltay; // d->dir == 3 ? deltay : planeht - deltay;// ycoordinate where the two frames meet
			
				// the smaller the frame size (reduced) more is taper	
				int taper1 = ((planeht-meeth)* tapermax)/planeht;
				int taper2 = tapermax-taper1;
			
				if ( fi->sampleType == stInteger && nbytes == 1)
				{
					uint8_t * dp = dstp;
			
						// blacken sides of image in the upper part of frame 

					for(int h = meeth; h >= 0; h--)
						for(int w = 0; w < ((meeth - h) * taper1) / meeth; w ++)
						{
							*(dp + h * pitch + w) = blackCol[plane];
							*(dp + h * pitch + (planewd - 1 - w) ) = blackCol[plane];
						}

							// blacken image in the lower part of frame

					for(int h = meeth; h < planeht; h++)
						for(int w = 0; w < ((h - meeth) * taper2) / (planeht - meeth); w++)
						{
							*(dp + h * pitch + w) = blackCol[plane];
							*(dp + h * pitch + (planewd - 1 - w) ) = blackCol[plane];
						}
				}	// integer nbits  8

				else if( fi->sampleType == stInteger && nbytes == 2)
				{
					uint16_t * dp = (uint16_t *)dstp;
			
						// blacken sides of image in the upper part of frame 

					for(int h = meeth; h >= 0; h --)
						for(int w = 0; w < ((meeth - h) * taper1) / meeth; w++)
						{
							*(dp + h * pitch + w) = (uint16_t)(blackCol[plane] << (nbits - 8));
							*(dp + h * pitch + (planewd-1-w) ) = (uint16_t)(blackCol[plane] << (nbits - 8));
						}

							// blacken image in the lower part of frame

					for(int h = meeth; h < planeht; h ++)
						for(int w = 0; w < ((h - meeth) * taper2) / (planeht - meeth); w++)
						{
							*(dp + h * pitch + w) = (uint16_t)(blackCol[plane] << (nbits - 8));
							*(dp + h * pitch + (planewd-1-w) ) = (uint16_t)(blackCol[plane] << (nbits - 8));
						}
				}	// integer 

				else if( fi->sampleType == stFloat)
				{
					float * dp = (float *)dstp;
			
						// blacken sides of image in the upper part of frame 

					for(int h = meeth; h >= 0; h --)
						for(int w = 0; w < ((meeth - h) * taper1) / meeth; w ++)
						{
							*(dp + h * pitch + w) = 0;
							*(dp + h * pitch + (planewd - 1 - w) ) = 0;
						}

							// blacken image in the lower part of frame

					for(int h = meeth; h < planeht; h ++)
						for(int w = 0; w < ((h - meeth) * taper2) / (planeht - meeth); w++)
						{
							*(dp + h * pitch + w) = 0;
							*(dp + h * pitch + (planewd-1-w) ) = 0;
						}
				}	// float
			
			} // dir = 1 or 3
			
		
		}	// plane

		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
	
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC spriteFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    SpriteData *d = (SpriteData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC spriteCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    SpriteData d;
    SpriteData *data;
    int err;
	//int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Sprite: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Sprite: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Sprite: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir = int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err)
		d.dir = 1;

	if(  d.dir < 1 || d.dir > 4 )
	{
		vsapi->setError(out, "Sprite: dir valid numbers 1 for up, 2 for right, 3 for and 4 for left");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}	

	
	
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (SpriteData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Sprite", spriteInit, spriteGetFrame, spriteFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "sprite", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Sprite", "clip:clip;clipb:clip;overlap:float:opt;dir:int:opt;", spriteCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
