/******************************************************************************
Plugin function for vapoursynth
The Trans.Swirl Filter creates a swirl over which transition occurs.
During first 1/5 of transition swirl grows on A clip next 1/5th losing color,
next 1/5 merges with B swirl next 1/5th  gains color last 1/5th B clip  emerges 
 as swirl subsides 

5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int dir;	//1 clockwise or 2 anti swirl
	int qr;	// quality of swrl. 1 fine to 5 coarse
	int vtransStart;
	int vtransEnd;
	int numFrames;		
	
	
}SwirlData;


//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC swirlInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    SwirlData *d = (SwirlData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	
}

/***************************************************************/	
static const VSFrameRef *VS_CC swirlGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     SwirlData *d = (SwirlData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd )
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
		int radmax = (int)sqrt( (float)( width * width + height * height));
		
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	
		// A frame swirl starts and expands full size from 0 during phase1
		// A frame swirl loses color gradually during phase2
		// A swirl merges with full grey B swirl during phase 3
		// B swirl gains color phase 4
		// B swirl reduces to reveal undisturbed B phase 5

		int phase1 = (d->overlap )/ 5;  
		int phase2 = 2 * phase1;	
		int phase3 = 3 * phase1;
		int phase4 = 4 * phase1;
		// phase5 is rest of overlap
		if( fi->numPlanes == 1)
		{
			// already grey image. so skip phases 2 and 4
			phase1 = phase2;
			phase3 = phase4;
		}

		
		

		const uint8_t * srpa[] = {NULL, NULL, NULL};
		const uint8_t * srpb[] = {NULL, NULL, NULL};
		uint8_t * dsp[] = {NULL, NULL, NULL};
		uint8_t * wp[] = {NULL, NULL, NULL};
		int pitches[] = {0, 0, 0};;

		int subH = (fi->colorFamily == cmYUV) ? fi->subSamplingH : 0;
		int subW = (fi->colorFamily == cmYUV) ? fi->subSamplingW : 0;

		VSFrameRef * dst = (bn < phase2 ) ? vsapi->copyFrame ( srca, core) : vsapi->copyFrame(srcb, core);
		
		
		// as all plane pointers are needed for the swirl we get them here
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcpa = vsapi->getReadPtr(srca, plane);
			const uint8_t * srcpb = vsapi->getReadPtr(srcb, plane);
			int stride = vsapi->getStride(srca, plane);
			uint8_t *  dstp = vsapi->getWritePtr(dst, plane);
			
		//	int ht = vsapi->getFrameHeight(srca, plane);
		//	int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / kb;	
			srpa[plane] = srcpa;
			srpb[plane] = srcpb;
			dsp[plane] = dstp;
			pitches[plane] = pitch;
		}

		if( bn <= phase1)
		{				
					// increase swirl radius  from  zero to max 
			int erad = ((bn + 1) * radmax) / phase1;
			int irad = 0;
			
			// on dst we have srca frame
			if (fi->sampleType == stInteger && nb == 8)
			{
				uint8_t grey = 1 << (nb - 1);
				uint8_t *dp = dsp[0];
				const uint8_t * sp = srpa[0];

				if ( fi->numPlanes == 1) 
				{
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
					// single plane swirl increase radius  to maximum by end of this ophase
						vsORingRotatePlane(sp, dp, rad, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);
					}

				}

				else // YUV RGB
				{
					uint8_t *dpU = dsp[1];
					const uint8_t * spU = srpa[1];
					uint8_t *dpV = dsp[2];
					const uint8_t * spV = srpa[2];
					// create swirl increasingly covering frame by end of phase1
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{

						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}										
										
				}					
			}

			else if (fi->sampleType == stInteger && nb > 8)
			{
				uint16_t grey = 1 << (nb - 1);
				uint16_t *dp =  (uint16_t *)dsp[0];
				const uint16_t * sp = ( const uint16_t *)srpa[0];

				if ( fi->numPlanes == 1) 
				{
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
					// single plane swirl increase radius  to maximum by end of this ophase
						vsORingRotatePlane(sp, dp, rad, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);
					}

				}

				else // YUV RGB
				{
					uint16_t *dpU = (uint16_t *)dsp[1];
					const uint16_t * spU = (const uint16_t *) srpa[1];
					uint16_t *dpV = (uint16_t *)dsp[2];
					const uint16_t * spV = (const uint16_t *) srpa[2];
					// create swirl increasingly covering frame by end of phase1
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{

						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}										
										
				}					
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = (float)(1 << (nb - 1));
				float *dp =  (float *)dsp[0];
				const float * sp = ( const float *)srpa[0];

				if ( fi->numPlanes == 1) 
				{
					
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
					// single plane swirl increase radius  to maximum by end of this ophase
						vsORingRotatePlane(sp, dp, rad, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);
					}
				}

				else // YUV or rgb
				{
					float *dpU = (float *)dsp[1];
					const float * spU = (const float *)srpa[1];
					float *dpV = (float *)dsp[2];
					const float * spV = (const float *)srpa[2];
					
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{

						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}
				}
			}
		}	// phase1

		else if (bn <= phase2)
		{
			// swirl loses color and becomes grey. Grey formats will no come here										
			if (fi->sampleType == stInteger && nb == 8)
			{
				uint8_t grey = 1 << (nb - 1);
				uint8_t *dp = dsp[0];
				uint8_t *dpU = dsp[1];
				uint8_t *dpV = dsp[2];
				const uint8_t * spa = srpa[0];
				const uint8_t * spaU = srpa[1];
				const uint8_t * spaV = srpa[2];				

				if (fi->colorFamily == cmRGB || fi->colorFamily == cmYUV)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spa,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spaU, spaV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}

					if (fi->colorFamily == cmRGB)
					{
						// gradually grey the swirl
						GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, phase2 - bn, bn - phase1);
					}

					else	// YUV
					{
						ColorMergePlane(dpU, dpU, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
						ColorMergePlane(dpV, dpV, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
					}
				}
				else	//  Grey
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						vsORingRotatePlane(spa, dp, rad, width / 2, height / 2, height, width,
										width / 2, height / 2, height, width,
										pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);					
					}
				}
			}

			else if (fi->sampleType == stInteger && nb > 8)
			{
				uint16_t grey = 1 << (nb - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * spa = (const uint16_t *)srpa[0];
				const uint16_t * spb = (const uint16_t *)srpb[0];
				uint16_t *dpU = (uint16_t *)dsp[1];					
				uint16_t *dpV = (uint16_t *)dsp[2];
				const uint16_t * spaU = (const uint16_t *)srpa[1];
				const uint16_t * spaV = (const uint16_t *)srpa[2];
				const uint16_t * spbU = (const uint16_t *)srpb[1];					
				const uint16_t * spbV = (const uint16_t *)srpb[2];

				if (fi->colorFamily == cmRGB || fi->colorFamily == cmYUV)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spa,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spaU, spaV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}

					if (fi->colorFamily == cmRGB)
					{
						// gradually grey the swirl
						GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, phase2 - bn, bn - phase1);
					}

					else	// YUV
					{
						ColorMergePlane(dpU, dpU, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
						ColorMergePlane(dpV, dpV, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
					}
				}
				else	//  Grey
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						vsORingRotatePlane(spa, dp, rad, width / 2, height / 2, height, width,
										width / 2, height / 2, height, width,
										pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);					
					}
				}
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * spa = (const float *)srpa[0];
				const float * spb = (const float *)srpb[0];
				float *dpU = (float *)dsp[1];					
				float *dpV = (float *)dsp[2];
				const float * spaU = (const float *)srpa[1];
				const float * spaV = (const float *)srpa[2];
				const float * spbU = (const float *)srpb[1];					
				const float * spbV = (const float *)srpb[2];
				
				if (fi->colorFamily == cmRGB || fi->colorFamily == cmYUV)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spa,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spaU, spaV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}

					if (fi->colorFamily == cmRGB)
					{
						// gradually grey the swirl
						GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, phase2 - bn, bn - phase1);
					}

					else	// YUV
					{
						ColorMergePlane(dpU, dpU, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
						ColorMergePlane(dpV, dpV, pitches[1], width >> subW, height >> subH, 
									phase2 - bn, bn - phase1, grey);
					}
				}
				else	//  Grey
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						vsORingRotatePlane(spa, dp, rad, width / 2, height / 2, height, width,
										width / 2, height / 2, height, width,
										pitches[0],  pitches[0], d->dir * (4 * bn + rad), d->qr);					
					}
				}
			}
		}	// phase2 

		else if (bn <= phase3)
		{
			// grey swirl of A  merge with grey swirl of B finally becomes grey swirl of B										
			if (fi->sampleType == stInteger && nb == 8)
			{
				uint8_t grey = 1 << (nb - 1);
				uint8_t *dp = dsp[0];
				uint8_t *dpU = dsp[1];
				uint8_t *dpV = dsp[2];
				const uint8_t * spa = srpa[0];
				const uint8_t * spaU = srpa[1];
				const uint8_t * spaV = srpa[2];
				const uint8_t * spb = srpb[0];
				const uint8_t * spbU = srpb[1];
				const uint8_t * spbV = srpb[2];

				if ( fi->colorFamily == cmYUV)
				{					
					// grey u and V planes of dst					
					ColorPlane(dpU,  pitches[1], width >> subW, height >> subH, grey);
					ColorPlane(dpV,  pitches[1], width >> subW, height >> subH, grey);
				}
				if ( fi->colorFamily != cmRGB)
				{
						// create swirl of Y plane
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spa, spb, dp,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
				}
				else if( fi->colorFamily == cmRGB)
				{
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spaU, spbU, dpU,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
					GreyRGBFrame(dp, dpU, dpV, dp, dpU, dpV,
								pitches[0], width, height);
				}
				
			}

			else if (fi->sampleType == stInteger && nb > 8)
			{
				uint16_t grey = 1 << (nb - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * spa = (const uint16_t *)srpa[0];
				const uint16_t * spb = (const uint16_t *)srpb[0];
				uint16_t *dpU = (uint16_t *)dsp[1];					
				uint16_t *dpV = (uint16_t *)dsp[2];
				const uint16_t * spaU = (const uint16_t *)srpa[1];
				const uint16_t * spaV = (const uint16_t *)srpa[2];
				const uint16_t * spbU = (const uint16_t *)srpb[1];					
				const uint16_t * spbV = (const uint16_t *)srpb[2];

				if ( fi->colorFamily == cmYUV)
				{					
					// grey u and V planes of dst					
					ColorPlane(dpU,  pitches[1], width >> subW, height >> subH, grey);
					ColorPlane(dpV,  pitches[1], width >> subW, height >> subH, grey);
				}
				if ( fi->colorFamily != cmRGB)
				{
						// create swirl of Y plane
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spa, spb, dp,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
				}
				else if( fi->colorFamily == cmRGB)
				{
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spaU, spbU, dpU,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
					GreyRGBFrame(dp, dpU, dpV, dp, dpU, dpV,
								pitches[0], width, height);
				}
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * spa = (const float *)srpa[0];
				const float * spb = (const float *)srpb[0];
				float *dpU = (float *)dsp[1];					
				float *dpV = (float *)dsp[2];
				const float * spaU = (const float *)srpa[1];
				const float * spaV = (const float *)srpa[2];
				const float * spbU = (const float *)srpb[1];					
				const float * spbV = (const float *)srpb[2];

				if ( fi->colorFamily == cmYUV)
				{					
					// grey u and V planes of dst					
					ColorPlane(dpU,  pitches[1], width >> subW, height >> subH, grey);
					ColorPlane(dpV,  pitches[1], width >> subW, height >> subH, grey);
				}
				if ( fi->colorFamily != cmRGB)
				{
						// create swirl of Y plane
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spa, spb, dp,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
				}
				else if( fi->colorFamily == cmRGB)
				{
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// merge grey swirls. gradually A becoming B
						vsMergeORingRotatePlane(spaU, spbU, dpU,
											rad, width / 2, height / 2, height, width, pitches[0], 
											d->dir * (4 * bn + rad), d->qr, phase3 - bn, bn - phase2);
						
					}
					GreyRGBFrame(dp, dpU, dpV, dp, dpU, dpV,
								pitches[0], width, height);
				}
			}
		}// phase3
		
		else if ( bn <= phase4)
		{
			// we should/ will not be coming here for single plane format
			// in this phase the grey B swirl regains color and becomes color B swirl
			if (fi->sampleType == stInteger && nb == 8)
			{
				uint8_t grey = 1 << (nb - 1);
				uint8_t *dp = dsp[0];
				uint8_t *dpU = dsp[1];
				uint8_t *dpV = dsp[2];				
				const uint8_t * spb = srpb[0];
				const uint8_t * spbU = srpb[1];
				const uint8_t * spbV = srpb[2];				

				if (fi->colorFamily == cmRGB)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}
						// gradually un grey the swirl
					GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, bn - phase3, phase4 - bn );
				}
				else if (fi->colorFamily == cmYUV)	// YUV 
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);					
					}
				
					// un grey gradually UV planes
					ColorMergePlane(dpU, spbU, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn,  grey);
					ColorMergePlane(dpV, spbV, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn,  grey);
				}
					
			}

			else if (fi->sampleType == stInteger && nb > 8)
			{
				uint16_t grey = 1 << (nb - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * spb = (const uint16_t *)srpb[0];
				uint16_t *dpU = (uint16_t *)dsp[1];					
				uint16_t *dpV = (uint16_t *)dsp[2];
				const uint16_t * spbU = (const uint16_t *)srpb[1];					
				const uint16_t * spbV = (const uint16_t *)srpb[2];

				if (fi->colorFamily == cmRGB)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}
						// gradually un grey the swirl
					GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, bn - phase3, phase4 - bn);
				}
				else	// YUV or Grey
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);					
					}
				}
				if ( fi->colorFamily == cmYUV)
				{
					// un grey gradually UV planes
					ColorMergePlane(dpU, spbU, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn, grey);
					ColorMergePlane(dpV, spbV, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn, grey);
				}
			}

			else if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * spb = (const float *)srpb[0];
				float *dpU = (float *)dsp[1];					
				float *dpV = (float *)dsp[2];
				const float * spbU = (const float *)srpb[1];					
				const float * spbV = (const float *)srpb[2];
					
				if (fi->colorFamily == cmRGB)
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);						
					}
						// gradually un grey the swirl
					GreyMergeRGBFrame(dp, dpU, dpV,  dp, dpU, dpV,
									pitches[0], width, height, bn - phase3, phase4 - bn);
				}
				else	// YUV or Grey
				{					
					for( int rad = d->qr; rad < radmax; rad += d->qr)
					{
						// single plane full A swirl 
						// create a full frame swirl
						vsORingRotateYUVPlanes(spb,dp,rad, width / 2, height / 2, height, width,
															width / 2, height / 2, height, width,
												pitches[0], pitches[0], d->dir * (4 * bn + rad), d->qr,
													spbU, spbV, pitches[1], dpU, dpV, pitches[1], subH, subW);					
					}
				}
				if ( fi->colorFamily == cmYUV)
				{
					// un grey gradually UV planes
					ColorMergePlane(dpU, spbU, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn, grey);
					ColorMergePlane(dpV, spbV, pitches[1], width >> subW, height >> subH, 
									bn - phase3, phase4 - bn, grey);
				}
			}
		} // end of phase4

		else if (bn > phase4)	// last phase
		{
			// in this phase the b frame swirl which completely covered frame subsides and reveals B frame		
			int erad = ( (d->overlap - bn ) * radmax) / (d->overlap - phase4);
			 
			
			if (fi->sampleType == stInteger && nb == 8)
			{
				uint8_t grey = 1 << (nb - 1);
				uint8_t *dp = dsp[0];
				const uint8_t * sp = srpb[0];

				if ( fi->numPlanes == 1) 
				{
					// y8 format
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{						
						
						// single plane B swirl increase thickness  to full by end of this phase
						vsORingRotatePlane(sp, dp, radmax, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr);
					}
				}

				else // YUV or RGB
				{
					uint8_t *dpU = dsp[1];
					const uint8_t * spU = srpb[1];
					uint8_t *dpV = dsp[2];
					const uint8_t * spV = srpb[2];
					
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{

						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}
				}
			}

			if (fi->sampleType == stInteger && nb > 8)
			{
				uint16_t grey = 1 << (nb - 1);
				uint16_t *dp = (uint16_t *)dsp[0];
				const uint16_t * sp = (const uint16_t *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
				
					// y8 format
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
						// single plane B swirl increase thickness  to full by end of this phase
						vsORingRotatePlane(sp, dp, radmax, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr);
					}
				}

				else // YUV or RGB
				{
					uint16_t *dpU = (uint16_t *)dsp[1];
					const uint16_t * spU = (const uint16_t *)srpb[1];
					uint16_t *dpV = (uint16_t *)dsp[2];
					const uint16_t * spV = (const uint16_t *)srpb[2];
					//  
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}
				}
			}

			if (fi->sampleType == stFloat)
			{
				float grey = 0.0f;
				float *dp = (float *)dsp[0];
				const float * sp = (const float *)srpb[0];

				if ( fi->numPlanes == 1) 
				{
				
					// y8 format
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
						// single plane B swirl increase thickness  to full by end of this phase
						vsORingRotatePlane(sp, dp, radmax, width / 2, height / 2, height, width,
											width / 2, height / 2, height, width,
											pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr);
					}
				}

				else // YUV or RGB
				{
					float *dpU = (float *)dsp[1];
					const float * spU = (const float *)srpb[1];
					float *dpV = (float *)dsp[2];
					const float * spV = (const float *)srpb[2];
					//  reduces  radius
					for( int rad = d->qr; rad < erad; rad += d->qr)
					{
						vsORingRotateYUVPlanes(sp, dp, rad, width / 2, height / 2, height, width,
												width / 2, height / 2, height, width,
												pitches[0],  pitches[0], d->dir *  (4 * bn + rad), d->qr,
												spU, spV, pitches[1], dpU, dpV, pitches[1],											
												subH, subW);
					}
				}
			}
		}		
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC swirlFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    SwirlData *d = (SwirlData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC swirlCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    SwirlData d;
    SwirlData *data;
    int err;
	//int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Swirl: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Swirl: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)( ( overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Swirl: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.dir =  int64ToIntS(vsapi->propGetInt(in, "dir", 0, &err));
	if( err )
		d.dir = 1;
	else if (d.dir != 1 && d.dir != -1)
	{
		vsapi->setError(out, "Swirl: dir swirl rotation can have values of -1 or 1 only ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}

	
	d.qr =  int64ToIntS(vsapi->propGetInt(in, "qr", 0, &err));
	if( err )
		d.qr = 3;
	else if (d.qr < 1 || d.qr > 5)
	{
		vsapi->setError(out, "Swirl: quality of ripple qr must be between 1: fine to 5 : coarse ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	d.qr *= 4; 
		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (SwirlData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Swirl", swirlInit, swirlGetFrame, swirlFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "swirl", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Swirl", "clip:clip;clipb:clip;overlap:float:opt;qr:int:opt;dir:int:opt;", swirlCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
