/******************************************************************************
Trans.VenetianBlinds filter function for plugin for vapoursynth by V.C.Mohan
The A Clip  unmasks a stationary B Clip  in the pattern specified.
 The clips should have identical lengths, height, width and same pixel type 

5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 
*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int style;	//1 vertical  2 horizontal 3 diamonds 4 checkers
	int slatwidth;	// 
	int vtransStart;
	int vtransEnd;
	int numFrames;		
	bool open;
	
}VenitianBlindsData;
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC venitianblindsInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    VenitianBlindsData *d = (VenitianBlindsData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);

	
}

/***************************************************************/	
static const VSFrameRef *VS_CC venitianblindsGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     VenitianBlindsData *d = (VenitianBlindsData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
				
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		// number of slats along width / height
		int NWindowh = height / d->slatwidth;
		int NWindoww =  width / d->slatwidth;
		int hstart = (height % NWindowh) / 2;
		int wstart = (width % NWindoww) / 2;
		// current slat width
		int dwidth=(d->slatwidth * (bn + 1) ) / (d->overlap + 1);
		int dheight = (int)(0.8 * dwidth);
		dwidth = dwidth & 0xfffffffe;
		dheight = dheight & 0xfffffffe;

		if(dwidth < 2 || dheight < 2)
		{
			vsapi->freeFrame(srcb);
			return srca;
		}

		VSFrameRef * dst = vsapi->copyFrame ( srcb, core);
		const VSFrameRef *src = srca;

		if(! d->open)
		{
			dst = vsapi->copyFrame(srca, core);
			src = srcb;
			dwidth = d->slatwidth - dwidth;
			dheight = d->slatwidth - dheight;
		}

		
		// 
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcp = vsapi->getReadPtr(src, plane);
			
			int stride = vsapi->getStride(src, plane);
			uint8_t * dstp = vsapi->getWritePtr(dst, plane);
			int ht = vsapi->getFrameHeight(srca, plane);
			int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / kb;
			
			int subh = plane == 0 ? 0 : fi->colorFamily == cmYUV ? fi->subSamplingH : 0 ;
			int subw = plane == 0 ? 0 : fi->colorFamily == cmYUV ? fi->subSamplingW : 0 ;

			if ( fi->sampleType == stInteger && nb == 8)
			{
				const uint8_t * sp = (const uint8_t * )srcp;
				uint8_t * dp = (uint8_t * ) dstp;

				// applies to all video formats RGB, YUY2, Planar
			
				if(d->style == 2)		// horizontal slats
				{
					for(int i = 0; i < NWindowh; i ++)
					{
						int offset = (i * ((d->slatwidth +  hstart) >> subh) ) * pitch;
						const uint8_t * srp = sp + offset;
						uint8_t * dsp = dp + offset;

						for(int h = 0; h < dwidth >> subh; h ++)
						{							
							for(int w = 0; w < wd; w ++)
							{
								dsp[ w] = srp[w];
							}
							dsp += pitch;
							srp += pitch;
						}
					}
					
				}

				else if(d->style == 1)		// vertical slats
				{					
					for(int i = 0; i < NWindoww; i ++)
					{
						int offset = (i * ((d->slatwidth +  wstart) >> subw) );

						for(int w = 0; w < dwidth >> subw; w ++)
						{							
							const uint8_t * srp = sp + offset + w;
							uint8_t * dsp = dp + offset + w;

							for(int h = 0; h < ht; h ++)
							{
									dsp[0] = srp[0];
									dsp += pitch;
									srp += pitch;
							}
						}
					}
				
				}

				else if(d->style == 4)		// checkers pattern
				{
					for(int i = 0; i < NWindowh + 1; i ++)

						for(int h = 0; h < dwidth >> subh; h ++)

							for(int k = 0; k < NWindoww + 1; k ++)

								for(int w = 0; w < dwidth >> subw; w ++)

									if(i * (d->slatwidth >> subh) + h < ht && k * (d->slatwidth >> subw) + w < wd)										

											dp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )]

												 = sp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )];
			
				}
			
				else	//if(d->style == 3) diamonds pattern
				{
					int dwd = dwidth >> subw;
					int dht = dheight >> subh;

					for(int e = 0; e < 2; e ++)	// as diamonds are staggered

						for(int i = (int)(-e * 0.8 * (d->slatwidth >> subh));i < ht; i += (int)(1.6 * (d->slatwidth >> subh)))

							for(int k = -e * (d->slatwidth >> subw ) / 2; k < wd; k += d->slatwidth >> subw)

								for(int h = 0; h < dht; h ++)
									
									for(int w = 0; w < ( (dht - h) * dwd) / dht / 2; w ++)										
									{
										if(i + h >= 0 && i + h < ht)
										{
											if( k + w >= 0 && k + w < wd)													
													
												*(dp + (i + h) * pitch + (k + w) )

													= * (sp + (i + h) * pitch + (k + w) );

											if( k - w >= 0 && k - w < wd)

												*(dp + (i + h) * pitch + (k - w) )

													= * (sp + (i + h) * pitch + (k - w) );
										}

										if(i - h >= 0 && i - h < ht)
										{
											if( k - w >= 0 && k - w < wd)
													
												*(dp + (i - h) * pitch + (k - w) )

													= * (sp + (i - h) * pitch + (k - w) );

											if(k + w >= 0 && ( k + w) < wd)

												*(dp + (i - h) * pitch + (k + w) )

													= * (sp + (i - h) * pitch + (k + w) );
										}
									
									}

				}
			}	// integer 8 bits		

			else if ( fi->sampleType == stInteger && nb > 8)
			{
				const uint16_t * sp = (const uint16_t * )srcp;
				uint16_t * dp = (uint16_t * ) dstp;

				// applies to all video formats RGB, YUY2, Planar
			
				if(d->style == 2)		// horizontal slats
				{
					for(int i = 0; i < NWindowh; i ++)
					{
						int offset = (i * ((d->slatwidth +  hstart) >> subh) ) * pitch;
						const uint16_t * srp = sp + offset;
						uint16_t * dsp = dp + offset;

						for(int h = 0; h < dwidth >> subh; h ++)
						{							
							for(int w = 0; w < wd; w ++)
							{
								dsp[ w] = srp[w];
							}
							dsp += pitch;
							srp += pitch;
						}
					}
					
				}

				else if(d->style == 1)		// vertical slats
				{					
					for(int i = 0; i < NWindoww; i ++)
					{
						int offset = (i * ((d->slatwidth +  wstart) >> subw) );

						for(int w = 0; w < dwidth >> subw; w ++)
						{							
							const uint16_t * srp = sp + offset + w;
							uint16_t * dsp = dp + offset + w;

							for(int h = 0; h < ht; h ++)
							{
									dsp[0] = srp[0];
									dsp += pitch;
									srp += pitch;
							}
						}
					}
				
				}

				else if(d->style == 4)		// checkers pattern
				{
					for(int i = 0; i < NWindowh + 1; i ++)

						for(int h = 0; h < dwidth >> subh; h ++)

							for(int k = 0; k < NWindoww + 1; k ++)

								for(int w = 0; w < dwidth >> subw; w ++)

									if(i * (d->slatwidth >> subh) + h < ht && k * (d->slatwidth >> subw) + w < wd)										

											dp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )]

												 = sp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )];
			
				}
			
				else	//if(d->style == 3) diamonds pattern
				{
					int dwd = dwidth >> subw;
					int dht = dheight >> subh;

					for(int e = 0; e < 2; e ++)	// as diamonds are staggered

						for(int i = (int)(-e * 0.8 * (d->slatwidth >> subh));i < ht; i += (int)(1.6 * (d->slatwidth >> subh)))

							for(int k = -e * (d->slatwidth >> subw ) / 2; k < wd; k += d->slatwidth >> subw)

								for(int h = 0; h < dht; h ++)
									
									for(int w = 0; w < ( (dht - h) * dwd) / dht / 2; w ++)										
									{
										if(i + h >= 0 && i + h < ht)
										{
											if( k + w >= 0 && k + w < wd)													
													
												*(dp + (i + h) * pitch + (k + w) )

													= * (sp + (i + h) * pitch + (k + w) );

											if( k - w >= 0 && k - w < wd)

												*(dp + (i + h) * pitch + (k - w) )

													= * (sp + (i + h) * pitch + (k - w) );
										}

										if(i - h >= 0 && i - h < ht)
										{
											if( k - w >= 0 && k - w < wd)
													
												*(dp + (i - h) * pitch + (k - w) )

													= * (sp + (i - h) * pitch + (k - w) );

											if(k + w >= 0 && ( k + w) < wd)

												*(dp + (i - h) * pitch + (k + w) )

													= * (sp + (i - h) * pitch + (k + w) );
										}
									
									}

				}
			}	// integer  > 8 bits

			else if ( fi->sampleType == stFloat)
			{
				const float * sp = (const float * )srcp;
				float * dp = (float * ) dstp;

				// applies to all video formats RGB, YUY2, Planar
			
				if(d->style == 2)		// horizontal slats
				{
					for(int i = 0; i < NWindowh; i ++)
					{
						int offset = (i * ((d->slatwidth +  hstart) >> subh) ) * pitch;
						const float * srp = sp + offset;
						float * dsp = dp + offset;

						for(int h = 0; h < dwidth >> subh; h ++)
						{							
							for(int w = 0; w < wd; w ++)
							{
								dsp[ w] = srp[w];
							}
							dsp += pitch;
							srp += pitch;
						}
					}
					
				}

				else if(d->style == 1)		// vertical slats
				{					
					for(int i = 0; i < NWindoww; i ++)
					{
						int offset = (i * ((d->slatwidth +  wstart) >> subw) );

						for(int w = 0; w < dwidth >> subw; w ++)
						{							
							const float * srp = sp + offset + w;
							float * dsp = dp + offset + w;

							for(int h = 0; h < ht; h ++)
							{
									dsp[0] = srp[0];
									dsp += pitch;
									srp += pitch;
							}
						}
					}
				
				}

				else if(d->style == 4)		// checkers pattern
				{
					for(int i = 0; i < NWindowh + 1; i ++)

						for(int h = 0; h < dwidth >> subh; h ++)

							for(int k = 0; k < NWindoww + 1; k ++)

								for(int w = 0; w < dwidth >> subw; w ++)

									if(i * (d->slatwidth >> subh) + h < ht && k * (d->slatwidth >> subw) + w < wd)										

											dp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )]

												 = sp[(i * (d->slatwidth >> subh) + h) * pitch + (k * (d->slatwidth >> subw) + w )];
			
				}
			
				else	//if(d->style == 3) diamonds pattern
				{
					int dwd = dwidth >> subw;
					int dht = dheight >> subh;

					for(int e = 0; e < 2; e ++)	// as diamonds are staggered

						for(int i = (int)(-e * 0.8 * (d->slatwidth >> subh));i < ht; i += (int)(1.6 * (d->slatwidth >> subh)))

							for(int k = -e * (d->slatwidth >> subw ) / 2; k < wd; k += d->slatwidth >> subw)

								for(int h = 0; h < dht; h ++)
									
									for(int w = 0; w < ( (dht - h) * dwd) / dht / 2; w ++)										
									{
										if(i + h >= 0 && i + h < ht)
										{
											if( k + w >= 0 && k + w < wd)													
													
												*(dp + (i + h) * pitch + (k + w) )

													= * (sp + (i + h) * pitch + (k + w) );

											if( k - w >= 0 && k - w < wd)

												*(dp + (i + h) * pitch + (k - w) )

													= * (sp + (i + h) * pitch + (k - w) );
										}

										if(i - h >= 0 && i - h < ht)
										{
											if( k - w >= 0 && k - w < wd)
													
												*(dp + (i - h) * pitch + (k - w) )

													= * (sp + (i - h) * pitch + (k - w) );

											if(k + w >= 0 && ( k + w) < wd)

												*(dp + (i - h) * pitch + (k + w) )

													= * (sp + (i - h) * pitch + (k + w) );
										}
									
									}

				}
			}	// float

		}		
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC venitianblindsFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    VenitianBlindsData *d = (VenitianBlindsData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC venitianblindsCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    VenitianBlindsData d;
    VenitianBlindsData *data;
    int err;
	int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "VenitianBlinds: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "VenitianBlinds: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "VenitianBlinds: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.style =  int64ToIntS(vsapi->propGetInt(in, "style", 0, &err));
	if( err )
		d.style = 1;
	else if (d.style < 1 || d.style > 4)
	{
		vsapi->setError(out, "VenitianBlinds: valid style values are 1 : vertical, 2: horizontal, 3: diamond, 4: chckers ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	if( d.vi.width < 40 || d.vi.height < 40)
	{
		vsapi->setError(out, "VenitianBlinds: Too small frame. frame size must be atleast 40 X 40 pixels");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
	d.slatwidth =  int64ToIntS(vsapi->propGetInt(in, "slatwidth", 0, &err));
	if( err )
	{
		d.slatwidth = d.vi.width < d.vi.height ? d.vi.width / 16 : d.vi.height / 16;
		
	}
	if (d.slatwidth < 8 || d.slatwidth > d.vi.width / 8 || d.slatwidth > d.vi.height / 8)
	{
		vsapi->setError(out, "VenitianBlinds: slatwidth must be between 8 and 1/8 of frame smaller dimension");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}

	temp = !! vsapi->propGetInt(in, "open", 0, &err);
	if( err )
	{
		d.open = true;
		
	}
	else if( temp < 0 || temp > 1)
	{
		vsapi->setError(out, "VenitianBlinds: open can be either value 1 for open or 0 for close ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	if(temp == 0)
		d.open = false;
	else
		d.open = true;

		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (VenitianBlindsData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "VenitianBlinds", venitianblindsInit, venitianblindsGetFrame, venitianblindsFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "venitianblinds", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("VenitianBlinds", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;slatwidth:int:opt;open:int:opt;", venitianblindsCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
