/******************************************************************************
Trans.Weave filter function for plugin for vapoursynth by V.C.Mohan
The B Clip  covers  A Clip either as slats moving from opposite, or JigSaw in hor & ver.
 The clips should have identical lengths, height, width and same pixel type 

  5 Aug 2015  3 june 2021
 Author V.C.Mohan

  (C) <2015 - 2021>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    see <http://www.gnu.org/licenses/>.
	For details of how to contact author see <http://www.avisynth.org/vcmohan> 

*************************************************************************************************/  
typedef struct {
    VSNodeRef *node[2];
    VSVideoInfo vi;
	VSVideoInfo bvi;
	
    int overlap; // transition length in frames
	int style;	//1 : vertical, 2: horizontal, 3: jigsaw, 4: weave 
	int slatwidth;	// 
	int vtransStart;
	int vtransEnd;
	int numFrames;		
	bool open;
	
}WeaveData;
//*************************************************
// This function is called immediately after vsapi->createFilter(). This is the only place where the video
// properties may be set. In this case we simply use the same as the input clip. You may pass an array
// of VSVideoInfo if the filter has more than one output, like rgb+alpha as two separate clips.
static void VS_CC weaveInit(VSMap *in, VSMap *out, void **instanceData, VSNode *node, VSCore *core, const VSAPI *vsapi) 
{
    WeaveData *d = (WeaveData *) * instanceData;	
	d->vi.numFrames = d->numFrames;
    vsapi->setVideoInfo(&d->vi, 1, node);
	d->slatwidth = 16;
	
}

/***************************************************************/	
static const VSFrameRef *VS_CC weaveGetFrame(int n, int activationReason, void **instanceData, 
			void **frameData, VSFrameContext *frameCtx, VSCore *core, const VSAPI *vsapi) 
{
     WeaveData *d = (WeaveData *) * instanceData;

    if (activationReason == arInitial) 
	{
        // Request the source frame on the first call
		if( n <= d->vtransEnd)
			vsapi->requestFrameFilter(n, d->node[0], frameCtx);
		if ( n >= d->vtransStart)
			vsapi->requestFrameFilter(n - d->vtransStart, d->node[1], frameCtx);

    } 
	else if (activationReason == arAllFramesReady) 
	{
		if( n < d-> vtransStart)
	
			return (vsapi->getFrameFilter(n, d->node[0], frameCtx) );
		else if ( n > d-> vtransEnd)
			return (vsapi->getFrameFilter(n - d->vtransStart, d->node[1], frameCtx));

		int bn = n - d-> vtransStart;// corresponding bframe number
        const VSFrameRef *srca = vsapi->getFrameFilter(n, d->node[0], frameCtx);
		const VSFrameRef *srcb = vsapi->getFrameFilter(bn, d->node[1], frameCtx);
        // The reason we query this on a per frame basis is because we want our filter
        // to accept clips with varying dimensions. If we reject such content using d->vi
        // would be better.
        const VSFormat *fi = d->vi.format;

        int height = vsapi->getFrameHeight(srca, 0);
        int width = vsapi->getFrameWidth(srca, 0);
				
		int kb = fi->bytesPerSample;
		int nb = fi->bitsPerSample;	

		// current slat width		
		int deltax = (width * (bn + 1) ) / (d->overlap + 1);
		int deltay = (height * (bn + 1) ) / (d->overlap + 1);
		
		if(deltax <4 || deltay<4)
		{
			vsapi->freeFrame(srcb);
			return srca;
		}
		else if(deltax > width - 4 || deltay > height - 4)
		{
			vsapi->freeFrame(srca);
			return srcb;
		}

		VSFrameRef * dst = vsapi->copyFrame ( srca, core);
		const VSFrameRef *src = srcb;		
		// 
		for ( int plane = 0; plane < fi->numPlanes; plane ++)
		{
			const uint8_t * srcp = vsapi->getReadPtr(src, plane);
			
			int stride = vsapi->getStride(src, plane);
			uint8_t * dstp = vsapi->getWritePtr(dst, plane);
			int ht = vsapi->getFrameHeight(srca, plane);
			int wd = vsapi->getFrameWidth(srca, plane);
			int pitch = stride / kb;
			
			int subH = plane == 0 ? 0 : fi->colorFamily == cmYUV ? fi->subSamplingH : 0 ;
			int subW = plane == 0 ? 0 : fi->colorFamily == cmYUV ? fi->subSamplingW : 0 ;
			int andH = (1 << subH) -1;
			int andW = (1 << subW) -1;

			int deltax1 = deltax >>subW;
			int deltay1 = deltay >> subH;
			int slath = d->slatwidth >> subH;
			int slatw = d->slatwidth >> subW;

			if ( fi->sampleType == stInteger && nb == 8)
			{
				const uint8_t * sp = (const uint8_t * )srcp;
				uint8_t * dp = (uint8_t * ) dstp;
				
				
				if(d->style == 2)	// horizontal
				{
					for(int sh=0; sh<ht-slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++)
									*(dp+(sh+sw)* pitch+ w )
										=*(sp+(sh+sw)* pitch+ (wd-deltax1+w) );

					for( int sh= slath; sh<ht- slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++) 
									*(dp+(sh+sw)* pitch+ (wd-deltax1+w) )
											=*(sp+(sh+sw)* pitch+ w );


				
				}

				else if(d->style == 1)	// vertical
				{
					for(int ws=0; ws<=wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++)
									*(dp+h* pitch+ (ws+sw))
										=*(sp+(ht-deltay1+h)* pitch+ (ws+sw) );

					for(int ws= slatw; ws<wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++) 
									*(dp+(ht-deltay1+h)* pitch+ (ws+sw) )
										=*(sp+h* pitch+ (ws+sw) );
					
				
				
				}

				else if(d->style == 3)	// jigsaw
				{

					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath -1; h++)
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
										
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );

				
				}

				else	//if(d->style == 4)	// weave
				{
				
						// fill the four quarters with four different quadrants of input
					int deltax1= (deltax & 0xfffffff0 ) >> subW;
					int deltay1= (deltay & 0xfffffff0) >> subH;
					
						
						// copy alternate squares from alternate strips
						// squares from horizontally moving blocks
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw=0;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw= slatw;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}

						// squares from vertically moving blocks. these should be adjacent to h moving
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=slath; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=0; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
				
				}

	
			}	// integer 8 bits

			if ( fi->sampleType == stInteger && nb > 8)
			{
				const uint16_t * sp = (const uint16_t * )srcp;
				uint16_t * dp = (uint16_t * ) dstp;
				
				
				if(d->style == 2)	// horizontal
				{
					for(int sh=0; sh<ht-slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++)
									*(dp+(sh+sw)* pitch+ w )
										=*(sp+(sh+sw)* pitch+ (wd-deltax1+w) );

					for( int sh= slath; sh<ht- slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++) 
									*(dp+(sh+sw)* pitch+ (wd-deltax1+w) )
											=*(sp+(sh+sw)* pitch+ w );


				
				}

				else if(d->style == 1)	// vertical
				{
					for(int ws=0; ws<=wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++)
									*(dp+h* pitch+ (ws+sw))
										=*(sp+(ht-deltay1+h)* pitch+ (ws+sw) );

					for(int ws= slatw; ws<wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++) 
									*(dp+(ht-deltay1+h)* pitch+ (ws+sw) )
										=*(sp+h* pitch+ (ws+sw) );
					
				
				
				}

				else if(d->style == 3)	// jigsaw
				{

					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath -1; h++)
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
										
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );

				
				}

				else	//if(d->style == 4)	// weave
				{
				
						// fill the four quarters with four different quadrants of input
					int deltax1= (deltax & 0xfffffff0 ) >> subW;
					int deltay1= (deltay & 0xfffffff0) >> subH;
					
						
						// copy alternate squares from alternate strips
						// squares from horizontally moving blocks
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw=0;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw= slatw;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}

						// squares from vertically moving blocks. these should be adjacent to h moving
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=slath; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=0; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
				
				}

	
			}	// integer > 8 bits

			if ( fi->sampleType == stFloat)
			{
				const float * sp = (const float * )srcp;
				float * dp = (float * ) dstp;
				
				
				if(d->style == 2)	// horizontal
				{
					for(int sh=0; sh<ht-slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++)
									*(dp+(sh+sw)* pitch+ w )
										=*(sp+(sh+sw)* pitch+ (wd-deltax1+w) );

					for( int sh= slath; sh<ht- slath; sh+=2* slath)
						for(int sw=0; sw< slatw; sw++)
							for(int w=0; w<deltax1; w++) 
									*(dp+(sh+sw)* pitch+ (wd-deltax1+w) )
											=*(sp+(sh+sw)* pitch+ w );


				
				}

				else if(d->style == 1)	// vertical
				{
					for(int ws=0; ws<=wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++)
									*(dp+h* pitch+ (ws+sw))
										=*(sp+(ht-deltay1+h)* pitch+ (ws+sw) );

					for(int ws= slatw; ws<wd- slatw; ws+=2* slatw)
						for(int sw=0; sw< slatw; sw++)
							for(int h=0; h<deltay1; h++) 
									*(dp+(ht-deltay1+h)* pitch+ (ws+sw) )
										=*(sp+h* pitch+ (ws+sw) );
					
				
				
				}

				else if(d->style == 3)	// jigsaw
				{

					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath -1; h++)
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) );
										
					for(int sh= slath; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw=0; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (wd-deltax1+sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );
					for(int sh=0; sh<deltay1- slath; sh+=2* slath)
						for(int h=0; h< slath-1; h++)							
							for(int sw= slatw; sw<deltax1- slatw; sw+=2* slatw)
								for(int w=0; w< slatw-1; w++) 
										*(dp+(ht-deltay1+sh+h)* pitch+ (sw+w) )
											=*(sp+(sh+h)* pitch+ (wd-deltax1+sw+w) );

				
				}

				else	//if(d->style == 4)	// weave
				{
				
						// fill the four quarters with four different quadrants of input
					int deltax1= (deltax & 0xfffffff0 ) >> subW;
					int deltay1= (deltay & 0xfffffff0) >> subH;
					
						
						// copy alternate squares from alternate strips
						// squares from horizontally moving blocks
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw=0;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}
					for(int sh=slath/2;sh<ht/2-slath; sh+=3*slath/2)
						for(int h=sh;h<sh+slath-1; h++)
							for(int sw= slatw;sw<deltax1-2* slatw;sw+=3* slatw)
								for(int w=sw;w<sw+2* slatw-1;w++)
									{
										*(dp+(h)*pitch+w)
											=*(sp+(h)*pitch+(wd-deltax1+w));
										*(dp+(ht/2+h)*pitch+(wd-deltax1+w))
											=*(sp+(ht/2+h)*pitch+w);
									}

						// squares from vertically moving blocks. these should be adjacent to h moving
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=slath; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
					for(int sw= slatw/2;sw<wd/2- slatw;sw+=3* slatw/2)
						for(int w=sw;w<sw+ slatw-1;w++)
							for(int sh=0; sh<deltay1-2*slath;sh+=3*slath)
								for(int h=sh; h<sh+2*slath-1;h++)
									{
										*(dp+h*pitch+(wd/2+w))
											=*(sp+(ht-deltay1+h)*pitch+(wd/2+w));
										*(dp+(ht-deltay1+h)*pitch+(w))
											=*(sp+h*pitch+(w));
									}
				
				}

	
			}	// float
		}
		
		vsapi->freeFrame(srca);
		vsapi->freeFrame(srcb);
		return dst;
	}
	return 0;	
}
//------------------------------------------------------------------
// Free all allocated data on filter destruction
static void VS_CC weaveFree(void *instanceData, VSCore *core, const VSAPI *vsapi) 
{
    WeaveData *d = (WeaveData *)instanceData;
    vsapi->freeNode(d->node[0]);
	vsapi->freeNode(d->node[1]);
	
    free(d);
}

//--------------------------------------------------------------------
// This function is responsible for validating arguments and creating a new filter
static void VS_CC weaveCreate(const VSMap *in, VSMap *out, void *userData, VSCore *core, const VSAPI *vsapi)
{
    WeaveData d;
    WeaveData *data;
    int err;
	//int temp;
    // Get a clip reference from the input arguments. This must be freed later.
    d.node[0] = vsapi->propGetNode(in, "clip", 0, 0);
    d.vi = *vsapi->getVideoInfo(d.node[0]);

    
    // vi->format can be 0 if the input clip can change format midstream.
    if (!isConstantFormat(&d.vi) )
	{
        vsapi->setError(out, "Weave: only constant format  input supported");
        vsapi->freeNode(d.node[0]);
        return;
    }
	d.node[1] = vsapi->propGetNode(in, "clipb", 0, 0);
    d.bvi = *vsapi->getVideoInfo(d.node[1]);    
    
    if (!isSameFormat(&d.vi, &d.bvi) )
	{
        vsapi->setError(out, "Weave: both clips must have identical formats ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
    }

	float overlap = (float)vsapi->propGetFloat(in, "overlap", 0, &err);
	if (err)
	{
		overlap = 1.0f;
	}
	
	d.overlap = (int)((  overlap * d.vi.fpsNum ) / d.vi.fpsDen) ;
	
	if ( d.overlap < 8 || d.vi.numFrames < d.overlap ||  d.bvi.numFrames < d.overlap )
	{
		vsapi->setError(out, "Weave: overlap must result in atleast 8 frames and clips must have at least frames equal to overlap ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	// set new length
	
	d.vtransStart = d.vi.numFrames - d.overlap;
	d.vtransEnd = d.vi.numFrames - 1;
	d.numFrames = d.vi.numFrames + d.bvi.numFrames - d.overlap;

	d.style =  int64ToIntS(vsapi->propGetInt(in, "style", 0, &err));
	if( err )
		d.style = 1;
	else if (d.style < 1 || d.style > 4)
	{
		vsapi->setError(out, "Weave: valid style values are 1 : vertical, 2: horizontal, 3: jigsaw, 4: weave ");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	if( d.vi.width < 40 || d.vi.height < 40)
	{
		vsapi->setError(out, "Weave: Too small frame. frame size must be atleast 40 X 40 pixels");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}
	
	d.slatwidth = int64ToIntS(vsapi->propGetInt(in, "slatwidth", 0, &err));
	if( err )
	{
		d.slatwidth = d.vi.width < d.vi.height ? d.vi.width / 16 : d.vi.height / 16;
		
	}
	if (d.slatwidth < 8 || d.slatwidth > d.vi.width / 8 || d.slatwidth > d.vi.height / 8)
	{
		vsapi->setError(out, "Weave: slatwidth must be between 8 and 1/8 of frame smaller dimension");
        vsapi->freeNode(d.node[0]);
		vsapi->freeNode(d.node[1]);
        return;
	}

	

		 // I usually keep the filter data struct on the stack and don't allocate it
    // until all the input validation is done.
    data = (WeaveData *)malloc(sizeof(d));
    *data = d;

    // Creates a new filter and returns a reference to it. Always pass on the in and out
    // arguments or unexpected things may happen. The name should be something that's
    // easy to connect to the filter, like its function name.
    // The three function pointers handle initialization, frame processing and filter destruction.
    // The filtermode is very important to get right as it controls how threading of the filter
    // is handled. In general you should only use fmParallel whenever possible. This is if you
    // need to modify no shared data at all when the filter is running.
    // For more complicated filters, fmParallelRequests is usually easier to achieve as it can
    // be prefetched in parallel but the actual processing is serialized.
    // The others can be considered special cases where fmSerial is useful to source filters and
    // fmUnordered is useful when a filter's state may change even when deciding which frames to
    // prefetch (such as a cache filter).
    // If your filter is really fast (such as a filter that only resorts frames) you should set the
    // nfNoCache flag to make the caching work smoother.
    vsapi->createFilter(in, out, "Weave", weaveInit, weaveGetFrame, weaveFree, fmParallel, 0, data, core);
}

//////////////////////////////////////////
// Init

// This is the entry point that is called when a plugin is loaded. You are only supposed
// to call the two provided functions here.
// configFunc sets the id, namespace, and long name of the plugin (the last 3 arguments
// never need to be changed for a normal plugin).
//
// id: Needs to be a "reverse" url and unique among all plugins.
// It is inspired by how android packages identify themselves.
// If you don't own a domain then make one up that's related
// to the plugin name.
//
// namespace: Should only use [a-z_] and not be too long.
//
// full name: Any name that describes the plugin nicely.
//
// registerFunc is called once for each function you want to register. Function names
// should be PascalCase. The argument string has this format:
// name:type; or name:type:flag1:flag2....;
// All argument name should be lowercase and only use [a-z_].
// The valid types are int,float,data,clip,frame,func. [] can be appended to allow arrays
// of type to be passed (numbers:int[])
// The available flags are opt, to make an argument optional, empty, which controls whether
// or not empty arrays are accepted
/*
VS_EXTERNAL_API(void) VapourSynthPluginInit(VSConfigPlugin configFunc, VSRegisterFunction registerFunc, VSPlugin *plugin)
{
    configFunc("in.trans.vcm", "weave", "VapourSynth Accord function", VAPOURSYNTH_API_VERSION, 1, plugin);
    registerFunc("Weave", "clip:clip;clipb:clip;overlap:float:opt;style:int:opt;slatwidth:int:opt;", weaveCreate, 0, plugin);
}
*/

//----------------------------------------------------------------------------------------------------------
